/*---------------------------------------------------------------------------
 * 
 * Synopsis
 * --------
 * Sending and receiving character streams.
 *
 * Author
 * ------
 * Raymond McKendall
 * September 1998 -- January 1999
 *
 *--------------------------------------------------------------------------*/

#include "comm.h"

#define SRFLAGS 0  /* Flag parameter for send() and recv(). */

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

int comm_send_chars(
   SOCKET sfd,         /* Socket to use.           */
   const char msg[],   /* Outgoing chars.          */
   int n_to_send       /* Number of chars to send. */
)
/*
 * Description
 * -----------
 * Sends initial n_to_send chars of msg[] over socket sfd.
 *
 *---------------------------------------------------------------------------*/
{
  int n_sent = 0;   /* Counts chars sent. */
 
#ifdef WIN32
  if (WSAIsBlocking()) 
  {
	  printf("WSAIsBlocking()==TRUE Pre\n");
  }
#endif

  while (n_to_send > 0) {
    n_sent = send(sfd, msg, n_to_send, SRFLAGS);
    //n_sent = write(sfd, msg, n_to_send);
#ifdef WIN32
	  if (WSAIsBlocking()) 
	  {
		  printf("WSAIsBlocking()==TRUE post\n");
	  }
#endif

    if (n_sent != SOCKET_ERROR) {
      n_to_send -= n_sent;
      msg       += n_sent;
    }
    else {
      comm_error("comm_send_chars: send() failed");
      return 1;
    }
  }
  
  return 0;
}

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

int comm_recv_chars(
   SOCKET sfd,     /* Socket to use.              */
   char msg[],     /* Receives incoming chars.    */
   int n_to_recv   /* Number of chars to receive. */
)
/*
 * Description
 * -----------
 * Receives n_to_recv chars into msg[] from socket sfd.
 *
 *---------------------------------------------------------------------------*/
{
  int n_received = 0;    /* Counts chars received. */

  while (n_to_recv > 0) {
    n_received = recv(sfd, msg, n_to_recv, SRFLAGS);
	
	printf("nr=%i n2r=%i\n",n_received,n_to_recv);

    if (n_received != SOCKET_ERROR) {
      n_to_recv -= n_received;
      msg       += n_received;
    }
    else {
      comm_error("comm_recv_chars: recv() failed");
      return 1;
    }
  }
  
  return 0;
}

/*---------------------------------------------------------------------------*/
/*---- Test Module ----------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

#ifdef TEST
#undef TEST

#define CheckRC(msg) {if(rc) {fprintf(stderr, "main: %s\n", msg); exit(1);}}

#define ECHO_PORT 7

int main(int argc, char *argv[])
{
  SOCKET sfd;          /* Socket to server.               */
  char* host;          /* Host name or address of server. */
  char msg_send[40];   /* Buffer for outgoing message.    */
  char msg_recv[40];   /* Buffer for incoming message.    */
  u_long size_send;    /* Size of outgoing message.       */
  u_long size_recv;    /* Size of incoming message.       */
  int rc;              /* Return code.                    */

  /*--- Get host from command line. ---*/
  argc--;
  if (argc == 1) {
    host = *++argv;
  }
  else {
    fprintf(stderr,"main: usage: %s <echo host>\n", argv[0]);
    exit(1);
  }

  /*--- Engage server ---*/
  rc= comm_engage_server(&sfd, host, ECHO_PORT);
  CheckRC("Could not open server");
  printf("Connected to \"%s\" on port %d.\n", host, ECHO_PORT);

  /*--- Send and receive chars. ---*/
  sprintf(msg_send, "Hello, World!");
  size_send = size_recv = strlen(msg_send)+1;

  rc= comm_send_chars(sfd, msg_send, size_send);   
  CheckRC("comm_send_chars() failed");
  printf("\nSent message:     \'%s\'\n", msg_send);

  rc= comm_recv_chars(sfd, msg_recv, size_recv);   
  CheckRC("comm_recv_chars() failed");
  printf("Received message: \'%s\'\n\n", msg_recv);

  /*--- Clean up ---*/
  comm_disengage_server(sfd);

  return 0;
}

#endif

/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

#undef SRFLAGS
