/*---------------------------------------------------------------------------
 * 
 * Synopsis
 * --------
 * Header file for comm API.
 *
 * Author
 * ------
 * Raymond McKendall
 * September 1998 -- February 1999
 *
 *--------------------------------------------------------------------------*/

#ifndef _COMM_H_
#define _COMM_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <sys/types.h>
#include "bsd-winsock.h"

/*--- Prototypes: client-utils.c ---*/

int comm_engage_server(
   SOCKET* sfd,        /* Socket to create for server.                  */
   const char host[],  /* Either "aaa.bbb.ccc.ddd" or "name.domain.net" */
   u_short port        /* Target server's port.                         */
);

void comm_disengage_server(
   SOCKET sfd    /* Socket to server. */
);

/*--- Prototypes: server-utils.c ---*/

int comm_start_service(
   SOCKET* sfd,    /* Socket for service.                 */
   u_short port,   /* Port for service.                   */
   int backlog     /* Pass-through parameter to listen(). */
);

void comm_stop_service(
   SOCKET sfd    /* Socket listening for service requests. */
);

/*--- Prototypes: char.c ---*/

int comm_send_chars(
   SOCKET sfd,         /* Socket to use.           */
   const char msg[],   /* Outgoing chars.          */
   int n_to_send       /* Number of chars to send. */
);

int comm_recv_chars(
   SOCKET sfd,     /* Socket to use.              */
   char msg[],     /* Receives incoming chars.    */
   int n_to_recv   /* Number of chars to receive. */
);

/*--- Prototypes: hello.c ---*/

int comm_send_hello(
   SOCKET sfd   /* Socket to use. */
);

int comm_recv_hello(
   SOCKET sfd   /* Socket to use. */
);

#ifdef __cplusplus
}
#endif

#endif