#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include "link-includes.h"
#include "constituents.h"
#include "command-line.h"
#define close closesocket
#define systemError GetLastError()

#include <errno.h>
#include <time.h>

#include <signal.h>
#include "comm.h"
extern FILE *extrakbfile;
extern int outputMode;
extern void toLower(char *str);
extern char *bin_search(char *skey, FILE *fp);
int isPronoun(char *name);

int linkcnt;
char kb_hbuf[4096];
char kb_tbuf[4096];
char kb_buf[4096];
char kb_tempb[4096];
char kb_trigrams[4096];
char kb_testreg[4096];

typedef struct rrule_struct rrule_node;
struct rrule_struct
{
	rrule_node *next;
	char pattern[64];
	char templ[256];
	
};

typedef struct rrule_rootstruct rrule_roots;
struct rrule_rootstruct
{
	rrule_node *ruleroots[256];	
};


rrule_roots rruleroot1;// = NULL;
rrule_roots rruleroot2;// = NULL;
//--------------------------------------
// maybe hash match on the last character in the pattern.
// i.e for most cd1 patterns end in a word class identifier
// like "* ss *.v", matching on the "v" at the end could speed
// things up. if the last char is "*" then the pattern has to be
// examined anyway, but if not then it can be safely skipped
//--------------------------------------
#define INBUFSIZE 511

void read_xml_cdrules(char *filename,rrule_roots *rr_root)
{
 // The cheesist non-validating XML processor imaginable
 // Assumes format of
 //  <rule> <pattern>pattern</pattern><head>head_template</head><define>?VARIBLE</define><body>body_template</body></rule>
 //  fixed format, looks at the closing tag, only one pattern, multiple head and body mix
 //  does look at head to generate <define>X</define> into "(#$thereExists ?X %h"

    FILE *fp;
	char inbuf[INBUFSIZE+1];
	char *sp;
	char *nxp;
	unsigned char c;
	unsigned char cm;
	int nxparsestate;
	rrule_node *newnode;

    if ((fp = dictopen("", filename, "r")) == NULL) return;

#define XM_START  1
#define XM_TAG    2
#define XM_CTAG   3
#define XM_GETS    4
#define XM_FINTAG 5
#define XM_ENDTAG 6

	nxparsestate=XM_START;
	nxp=&inbuf[0]; *nxp=0;
	while( feof(fp)==0) 
		
	{
		c = fgetc(fp); 
		switch(nxparsestate)
		{
		case XM_START:
			// looking for "<" in "<tag>"
			 if (c=='<') nxparsestate=XM_TAG;
			 break;

		case XM_TAG:
			// its either the front tag or the rear
			 if (c=='/') nxparsestate=XM_FINTAG; // "</foo>"
					else nxparsestate=XM_CTAG;   // "<foo>"
			// for <define>?EVENT%l</define> => "(#$thereExists ?EVENT%l %h"
			 if (c=='d') 
			 {
				 strcat(nxp,"(#$thereExists ");
				 nxp= &inbuf[0] + strlen(inbuf);

			 }
			 break;

		case XM_CTAG:
			if (c=='>') {
				//nxp=&inbuf[0]; *nxp=0; 
				nxparsestate=XM_GETS;}
					
			 break;

		case XM_GETS:
			if (c=='<'){ nxparsestate=XM_TAG; }// got the closing <
			else {  if(c>=0x20) {*nxp++=c;*nxp=0;} }
			 break;

		case XM_FINTAG: // process based on the closing 
			 // </pattern>
			if (c=='p') { *nxp++=':';*nxp=0;}

			// </include> to allow nesting
			if (c=='i') {
				read_xml_cdrules(inbuf,rr_root);			
				nxp=&inbuf[0]; *nxp=0;// restart the line
			}

			 // </head> or </body> or </define>
			if ((c=='b')||(c=='h')||(c=='d')) {
				*nxp++='%'; 
				if (c=='d') c='h';
				if (c=='b') c='n';
				*nxp++=c; 
				*nxp=0; 
			}

			 // </rule>
			if (c=='r') 
			{
				*nxp++=0; // terminate the virtual line
				sp=strstr(inbuf,":");
				if(sp!=NULL)
				{
					// Create a new chain
					// lines are in the format "<pattern>:<template>"
					newnode=xalloc(sizeof(rrule_node));
					*sp=0;
					strcpy(newnode->pattern,inbuf);
					sp++;
					strcpy(newnode->templ,sp);
					if ((strlen(newnode->pattern)>0)&&(strlen(newnode->templ)>0))
					{
						cm = newnode->pattern[0];
						cm = tolower(cm);
						newnode->next = rr_root->ruleroots[cm];
						rr_root->ruleroots[cm]=newnode;
					}
					else
					{
						xfree(newnode,sizeof(rrule_node));
					}
				}
			// restart the line
			nxp=&inbuf[0]; *nxp=0;
			}
			 nxparsestate=XM_ENDTAG;
			 break;

		case XM_ENDTAG:
			 if (c=='>') nxparsestate=XM_START;
			 break;
		}
	}
	fclose(fp);

}

void read_cdrules(char *filename,rrule_roots *rr_root)
{

    FILE *fp;
	char inbuf[INBUFSIZE+1];
	char *sp;
	unsigned char c;
	rrule_node *newnode;
	//rruleroot1=NULL;
	
    if ((fp = dictopen("", filename, "r")) == NULL) return;
	
	while (fgets(inbuf,INBUFSIZE-1,fp))
	{
		sp=strstr(inbuf,":");
		if(sp!=NULL)
		{
			// Create a new chain
			// lines are in the format "<pattern>:<template>"
			newnode=xalloc(sizeof(rrule_node));
			*sp=0;
			strcpy(newnode->pattern,inbuf);
			sp++;
			strcpy(newnode->templ,sp);
			if ((strlen(newnode->pattern)>0)&&(strlen(newnode->templ)>0))
			{
				c = newnode->pattern[0];
				c = tolower(c);
				newnode->next = rr_root->ruleroots[c];
				rr_root->ruleroots[c]=newnode;
			}
			else
			{
				xfree(newnode,sizeof(rrule_node));
			}
		}
	}
	fclose(fp);

}

void load_cdrules(void)
{
	unsigned char c;
	for( c = 0;c<255;c++)
	{
	rruleroot1.ruleroots[c]=NULL;
	rruleroot2.ruleroots[c]=NULL;
	}
	
	// One link patterns,mostly direct subject and objects
	read_xml_cdrules("rosex1.txt",  &rruleroot1); // Opencyc denotations
	read_xml_cdrules("rosex2.txt",  &rruleroot2); // Opencyc denotations

	return;

	read_cdrules("dnote1.txt",  &rruleroot1); // Opencyc denotations
	read_cdrules("FRNB1.txt",   &rruleroot1); // FrameNet
	read_cdrules("FRN1.txt",    &rruleroot1); // FrameNet
	read_cdrules("SVTR1.txt",   &rruleroot1); // OpenCyc Verbrules
	read_cdrules("cdrules1.txt",&rruleroot1); // Basic rules

	// Two link rules, mostly preposition phrase and linked verbs (ex. have seen)
	read_cdrules("LUZRX.txt",   &rruleroot2); // FrameNet Preposition Rules ??
	read_cdrules("SVTR2.txt",   &rruleroot2); // OpenCyc Verbrules
	read_cdrules("cdrules2.txt",&rruleroot2); // Basic rules



}

void toCyc(char *str)
{
  char *p;
  int i, len = strlen(str);
  for (i=0; i<len; i++){

	  //str[i] = tolower(str[i]);
	  if ((str[i]=='.')||(str[i]==0x27)) str[i]='-';
	  if (str[i]=='$') str[i+1]=toupper(str[i+1]);
  }
  
  // Things that should be lower case
  p=strstr(str,"#$Isa ");
  while(p!=NULL)
  {
   p[2]=tolower(p[2]);
  p=strstr(str,"#$Isa ");
  }

  p=strstr(str,"#$Genls ");
  while(p!=NULL)
  {
   p[2]=tolower(p[2]);
  p=strstr(str,"#$Genls ");
  }

  p=strstr(str,"(#$");
  while(p!=NULL)
  {
   p[3]=tolower(p[3]);
  p=strstr(p+3,"(#$");
  }

  //p=strstr(str,"(#$thereExists");
  //while(p!=NULL)
  //{
  // p[3]=toupper(p[3]);
  //p=strstr(str,"(#$thereExists");
  //}

}
/* Replace all occurences of 'from' with 'to' in 'str' */

char *strsubst(char *str, char from, char to)
{
    register char *p;

    for (p = str; *p != 0; ++p)
	if (*p == from)
	    *p = to;
    return str;
}

int matchlike( char *String, char *Pattern ) 
{ 
	// works like VB like pattern match
	// returns 1 if match 0 if no match
	// star is recursive
	// note: included a negative char set delimited by {}
	//       so to exclude noun and verb tags {nv}

    char   c, p, l; 
//	printf("mk ('%s','%s')\n",String,Pattern);

    for (; ;) { 
        switch (p = *Pattern++) { 
            case 0:                             // end of pattern 
                return *String ? 0 : 1;  // if end of string TRUE 

            case '*': 
                while (*String) {               // match zero or more char 
                    if (matchlike (String++, Pattern)==1) 
                        return 1; 
                } 
                return matchlike (String, Pattern); 

            case '?': 
                if (*String++ == 0)             // match any one char 
                    return 0;                   // not end of string 
                break; 

            case '[': 
                if ( (c = *String++) == 0)      // match char set 
                    return 0;                   // syntax 

                c = tolower(c); 
                l = 0; 
                while (p = *Pattern++) { 
                    if (p == ']')               // if end of char set, then 
                        return 0;           // no match found 

                    if (p == '-') {             // check a range of chars? 
                        p = *Pattern;           // get high limit of range 
                        if (p == 0  ||  p == ']') 
                            return 0;           // syntax 

                        if (c >= l  &&  c <= p) 
                            break;              // if in range, move on 
                    } 

                    l = p; 
                    if (c == p)                 // if char matches this element 
                        break;                  // move on 
                } 

                while (p  &&  p != ']')         // got a match in char set 
                    p = *Pattern++;             // skip to end of set 

                break;
				
            case '{':// Inverse char set. True if NOT in this set.
                if ( (c = *String++) == 0)      // match char set 
                    return 0;                   // syntax 

                c = tolower(c); 
                l = 0; 
                while (p = *Pattern++) { 
                    if (p == '}')               // if end of char set, then 
                        break;           // no match found 

                    if (p == '-') {             // check a range of chars? 
                        p = *Pattern;           // get high limit of range 
                        if (p == 0  ||  p == '}') 
                            return 0;           // syntax 

                        if (c >= l  &&  c <= p) 
                            return 0;              // if in range, move on 
                    } 

                    l = p; 
                    if (c == p)                 // if char matches this element 
                        return 0;                  // move on 
                } 

                while (p  &&  p != '}')         // got NO match in char set 
                    p = *Pattern++;             // skip to end of set 

                break; 

            default: 
                c = *String++; 
                if (tolower(c) != p)            // check for exact char 
                    return 0;                   // not a match 

                break; 
        } 
    } 
} 

char message[512];
char simplevec[512];
char pairvec[512];
char trivec[512];
char tempvec[512];
int cyclishQ;

extern void sock_print(SOCKET sock,char *message);
void nstep(SOCKET sock,int len, char *vec,char *lnode,char *rnode);

int cycsafe(char *message)
{
	if (matchlike(message,"*#$lgp-xp [?]sent #$left-wall #$[?!.]*")==1) return 0;
	if (matchlike(message,"*#$lgp-* [?]sent #$[?!.] #$right-wall*")==1) return 0;
	if (matchlike(message,"*#$lgp-* [?]sent #$* #$'s-p")==1) return 0;
	return 1;

}

void sock_template(SOCKET sock, char *templ, char *lnode,char *cnode, char *rnode,int LN,int CN, int RN)
{
	char *buf,*cpbuf,*sp;
	char c,c2;

	

	buf=&message[0];
	
	while (*templ!=0)
	{
	 c=*templ++;
	 switch (c)
	 {
	 case '@':

			cpbuf=lnode;
			while(*cpbuf!=0){*buf++ = *cpbuf++;};
			break;


	 case '%':
			c2=*templ++;
			switch (c2)
			{
			case  'L': // use the left word
				cpbuf=lnode;
				while(*cpbuf!=0){*buf++ = *cpbuf++;};
				break;
			case  'C': // use the left word
				cpbuf=cnode;
				while(*cpbuf!=0){*buf++ = *cpbuf++;};
				break;

			case  'R':// use the right node

				cpbuf=rnode;
				while(*cpbuf!=0){*buf++ = *cpbuf++;};
				break;

			case 'l': // use the left number
				cpbuf=buf;            // point to the append point
				sprintf(cpbuf,"%i",LN); // just print to the buffer
				while (*buf!=0) buf++;  // find the end
				break; 
			case 'c': // use the center number
				cpbuf=buf;            // point to the append point
				sprintf(cpbuf,"%i",CN); // just print to the buffer
				while (*buf!=0) buf++;  // find the end
				break; 

			case 'r': // use the right number
				cpbuf=buf;            // point to the append point
				sprintf(cpbuf,"%i",RN); // just print to the buffer
				while (*buf!=0) buf++;  // find the end
				break; 

			case 'n': // break to define end of term for body
				*buf++='\r';
				*buf++='\n';
				*buf++=0;
				// did we already print it ?
				sp=strstr(kb_buf,message);
				if (sp==NULL)
				{
					sprintf(kb_tempb,"%s %s",kb_buf,message);
					strcpy(kb_buf,kb_tempb);
					 
					//toCyc(message);
					//if (cycsafe(message)==1) sock_print(sock,message);
					// append to end and copy back
				}
				//reset buffer
				buf=&message[0];
				*buf=0;

				break;

			case 'h': // break to define end of term for header
				*buf++='\r';
				*buf++='\n';
				*buf++=0;
				// did we already print it ?
				sp=strstr(kb_hbuf,message);
				if (sp==NULL)
				{
					sprintf(kb_tempb,"%s %s",kb_hbuf,message);
					strcpy(kb_hbuf,kb_tempb);
					// add to tail
					sprintf(kb_tempb,"%s )",kb_tbuf);
					strcpy(kb_tbuf,kb_tempb);

					//toCyc(message);
					//if (cycsafe(message)==1) sock_print(sock,message);
					// append to end and copy back
				}
				//reset buffer
				buf=&message[0];
				*buf=0;

				break;

			case 'A':// add to accumulator TR
				*buf++=0;
				strcpy(kb_testreg,message);
				//reset buffer
				buf=&message[0];
				*buf=0;
				break;

			case 'G':// Guard, exit if TR NOT found in body
				*buf++=0;
				 sp=strstr(kb_buf,kb_testreg);
				 if (sp==NULL) return;
				//reset buffer
				buf=&message[0];
				*buf=0;
				break;

			case 'T':// Test, exit if TR IS found in body
				*buf++=0;
				 sp=strstr(kb_buf,kb_testreg);
				 if (sp!=NULL) return;
				//reset buffer
				buf=&message[0];
				*buf=0;
				break;

			case 'Q':// Query, exit if TR NOT found in message
				*buf++=0;
				 sp=strstr(message,kb_testreg);
				 if (sp==NULL) return;
				//reset buffer
				buf=&message[0];
				*buf=0;
				break;

			case 'S':// Safe, if TR IS found in message
				*buf++=0;
				 sp=strstr(message,kb_testreg);
				 if (sp!=NULL) return;
				//reset buffer
				buf=&message[0];
				*buf=0;
				break;

			case 'M':// Continue if match, exit if not match
				*buf++=0;
				if (matchlike(kb_testreg,message)==0) return;
				//reset buffer
				buf=&message[0];
				*buf=0;
				break;

			case 'N':// Continue if NOT match, exit if match
				*buf++=0;
				if (matchlike(kb_testreg,message)==1) return;
				//reset buffer
				buf=&message[0];
				*buf=0;
				break;

			default:
				templ--; // backup
				cpbuf=rnode;
				while(*cpbuf!=0){*buf++ = *cpbuf++;};
				//buf--;
				break;
				
			}
			break;

	 default: 
			*buf++=c;
			break;
	 }
	}
	*buf++=0; //terminate
	//toCyc(message);
	//if (cycsafe(message)==1) sock_print(sock,message);

//	sock_print(sock,message);

}

void addnote(SOCKET sock,char *llink,char *label,char *rlink)
{
		if (outputMode==1)
		{
			sprintf(message,"   <LI>%s--[%s]--%s</LI>\r\n", llink,label,rlink);
			sock_print(sock,message);
		}
		else
			if (cyclishQ==0)
			{
				sprintf(message,"<link><llink>%s</llink><label>%s</label><rlink>%s</rlink></link>\r\n",llink,label,rlink);
				sock_print(sock,message);
			}
			else
			{
				if (matchlike(llink,"cd-*")==1)

					sprintf(message,"     (#$CD-%s ?SENT #$%s-TheWord )\r\n",label,rlink);
				else
				if (matchlike(label,"isa")==1)
					sprintf(message,"     (#$isa ?SENT #$%s)\r\n",rlink);
					else
					sprintf(message,"     (#$LGP-%s ?SENT #$%s-TheWord #$%s-TheWord )\r\n",label,llink,rlink);

				toCyc(message);
				if (cycsafe(message)==1) sock_print(sock,message);
			}
	
		linkcnt++;
		//recursive check
			sprintf(tempvec,"%s %s %s",llink,label,rlink);
			toLower(tempvec);
			nstep(sock,1,tempvec,llink,rlink);

}

void addnote2(SOCKET sock,char *llink,char *label,char *rlink)
{
		if (outputMode==1)
		{
			sprintf(message,"   <LI>%s--[%s]--%s</LI>\r\n", llink,label,rlink);
			sock_print(sock,message);
		}
		else
			if (cyclishQ==0)
			{
				sprintf(message,"<link><llink>%s</llink><label>%s</label><rlink>%s</rlink></link>\r\n",llink,label,rlink);
				sock_print(sock,message);
			}
			else
			{
				sprintf(message,"     (#$LGP-%s ?SENT #$%s-TheWord #$%s-TheWord )\r\n",label,llink,rlink);
				toCyc(message);
				if (cycsafe(message)==1) sock_print(sock,message);
			}
		linkcnt++;
		//recursive check

}

void addnote3(SOCKET sock,char *module,char *resource,char *valence)
{
		if (outputMode==1)
		{
			sprintf(message,"   <LI>MOD: %s--[%s]--%s</LI>\r\n",module,resource,valence);
			sock_print(sock,message);
		}
		else
		{
			if (cyclishQ==0)
			{
			sprintf(message,"<connection><module>%s</module><resource>%s</resource><valence>%s</valence></connection>\r\n",module,resource,valence);
			sock_print(sock,message);
			sprintf(message,"<link><llink>%s</llink><label>%s</label><rlink>%s</rlink></link>\r\n",module,valence,resource);
			sock_print(sock,message);
			}
			else
			{
				sprintf(message,"     (#$LGP-%s ?SENT #$%s-TheWord #$%s-TheWord )\r\n",valence,module,resource);
				toCyc(message);
				if (cycsafe(message)==1) sock_print(sock,message);

			}
		}
		linkcnt++;
		//recursive check

}

void addnote_cycl1(SOCKET sock,char *llink,char *label,char *rlink)
{
		if (outputMode==1)
			sprintf(message,"   <LI>%s--[%s]--%s</LI>\r\n", llink,label,rlink);
		else
			sprintf(message,"    (#$LGP-%s ?SENT #$%s #$%s )\r\n",label,llink,rlink);
		sock_print(sock,message);
	
		linkcnt++;
		//recursive check

}


// generate a guess file with lines in the proper format
void guessrule(char *filename,char *vec,char *lnode,char *rnode)
{
	FILE *learnfp;
	int headerflag;
	char *leftterm;
	char *rightterm;

	learnfp=fopen(filename,"a");
	if (learnfp!=NULL)
	{
	fprintf(learnfp,"%s:",vec);
	
	headerflag=0;
	if (matchlike (lnode,"*.a")==1){ fprintf(learnfp,"(#$thereExists ?ITEM%%l %%h(#$is-attribute ?ITEM%%l \"%%L\")%%n"); leftterm="?ITEM%l"; headerflag=1;}
	if (matchlike (lnode,"*.v")==1){ fprintf(learnfp,"(#$thereExists ?EVENT%%l %%h(#$is-verb ?EVENT%%l \"%%L\")%%n"); leftterm="?EVENT%l"; headerflag=1;}
	if (matchlike (lnode,"*.n")==1){ fprintf(learnfp,"(#$thereExists ?ITEM%%l %%h(#$is-node ?ITEM%%l \"%%L\")%%n"); leftterm="?ITEM%l"; headerflag=1;}
	if (headerflag==0) {fprintf(learnfp,"(#$ThereExists ?ITEM%%l %%h(#$is-node ?ITEM%%l \"%%L\")%%n"); leftterm="?ITEM%l";}

	headerflag=0;
	if (matchlike (rnode,"*.a")==1){ fprintf(learnfp,"(#$thereExists ?ITEM%%r %%h(#$is-attribute ?ITEM%%r \"%%R\")%%n");rightterm="?ITEM%r"; headerflag=1; }
	if (matchlike (rnode,"*.v")==1){ fprintf(learnfp,"(#$thereExists ?EVENT%%r %%h(#$is-verb ?EVENT%%r \"%%R\")%%n");rightterm="?EVENT%r"; headerflag=1;}
	if (matchlike (rnode,"*.n")==1){ fprintf(learnfp,"(#$thereExists ?ITEM%%r %%h(#$is-node ?ITEM%%r \"%%R\")%%n");rightterm="?ITEM%r"; headerflag=1;}
	if (headerflag==0) {fprintf(learnfp,"(#$thereExists ?ITEM%%r %%h(#$is-node ?ITEM%%r \"%%R\")%%h");rightterm="?ITEM%r"; headerflag=1;}
	
	fprintf(learnfp,"(#$unknown-relation %s %s)%%n",leftterm,rightterm);
	fprintf(learnfp,"(#$unknown-relation %s %s)%%n",rightterm,leftterm);

	fprintf(learnfp,"\r\n");
	fclose(learnfp);
	}
}

void singlestepCD(SOCKET sock,char *vec,char *lnode,char *rnode,int LN, int RN)
{
	// General purpose processor
	int learnflg;
	unsigned char hashc;
	rrule_node *rp;

	learnflg=1;
	hashc=*vec;
	hashc=tolower(hashc);

	// in general, then wildcard "?" then wildcard "*"
		rp=rruleroot1.ruleroots[hashc];
		while (rp != NULL)
		{
			if (matchlike(vec,rp->pattern)==1)
			{
				sock_template(sock,rp->templ,lnode,"cnode",rnode,LN,0,RN);
				learnflg=0;
			}
			rp=rp->next;
		}
		// wildcard "?"
		hashc='?';
		rp=rruleroot1.ruleroots[hashc];
		while (rp != NULL)
		{
			if (matchlike(vec,rp->pattern)==1)
			{
				sock_template(sock,rp->templ,lnode,"cnode",rnode,LN,0,RN);
				learnflg=0;
			}
			rp=rp->next;
		}
		// wildcard '*'
		hashc='*';
		rp=rruleroot1.ruleroots[hashc];
		while (rp != NULL)
		{
			if (matchlike(vec,rp->pattern)==1)
			{
				sock_template(sock,rp->templ,lnode,"cnode",rnode,LN,0,RN);
				learnflg=0;
			}
			rp=rp->next;
		}
	
	if ((learnflg==1)&&(cyclishQ==1)) guessrule("cycR1.txt",vec,lnode,rnode);

	return;

	//------------------------------------
	// INGEST
	// breakfast.v, consume.v, devour.v, dine.v, down.v, drink.v, eat.v, 
	// feast.v, feed.v, gobble.v, gulp.v, gulp.n, guzzle.v, have.v, ingest.v, 
	// lap.v, lunch.v, munch.v, nibble.v, nosh.v, nurse.v, sip.v, sip.n, slurp.v, 
	// slurp.n, snack.v, sup.v, swig.v, swig.n
	//------------------------------------
	if (matchlike(vec,"*drank.v*")==1)
	{
		//if (matchlike(vec,"* drank.v")==1) addnote(sock,"sent","isa","CD-INGEST");
		//if (matchlike(vec,"* ss drank.v")==1) addnote(sock,"CD-INGEST","ACTOR",lnode);
		//if (matchlike(vec,"drank.v os *.n")==1) addnote(sock,"CD-INGEST","OBJECT",rnode);

		if (matchlike(vec,"* drank.v")==1) sock_template(sock,"(#$isa ?SENT CD-INGEST)",lnode,"cnode",rnode,LN,0,RN);
		if (matchlike(vec,"* ss drank.v")==1) sock_template(sock,"(#$CD-ACTOR ?SENT #$@-TheWord)",lnode,"cnode",rnode,LN,0,RN);
		if (matchlike(vec,"drank.v os *.n")==1) sock_template(sock,"(#$CD-OBJECT ?SENT #$%-TheWord)",lnode,"cnode",rnode,LN,0,RN);

	}

	
	if (matchlike(vec,"*consume.v*")==1)
	{
		if (matchlike(vec,"* consume.v")==1) addnote(sock,"sent","isa","CD-INGEST");
		if (matchlike(vec,"* ss consume.v")==1) addnote(sock,"CD-INGEST","ACTOR",lnode);
		if (matchlike(vec,"consume.v os *.n")==1) addnote(sock,"CD-INGEST","OBJECT",rnode);
	}


	if (matchlike(vec,"*ate.v*")==1)
	{
		if (matchlike(vec,"* ate.v")==1) addnote(sock,"sent","isa","CD-INGEST");
		if (matchlike(vec,"* ss ate.v")==1) addnote(sock,"CD-INGEST","ACTOR",lnode);
		if (matchlike(vec,"ate.v os *.n")==1) addnote(sock,"CD-INGEST","OBJECT",rnode);
	}

	if (matchlike(vec,"*smoked.v*")==1)
	{
		if (matchlike(vec,"* smoked.v")==1) addnote(sock,"sent","isa","CD-INGEST");
		if (matchlike(vec,"smoked.v os *.n")==1) addnote(sock,"CD-INGEST","OBJECT",rnode);
	}


	// contraversial call. "John had a kite."vs."John had a lobster."
	
	if (matchlike(vec,"*had.v*")==1)
	{
		if (matchlike(vec,"* had.v")==1) addnote(sock,"sent","isa","CD-INGEST");
		if (matchlike(vec,"* ss had.v")==1) addnote(sock,"CD-INGEST","ACTOR",lnode);
		if (matchlike(vec,"had.v os *")==1) addnote(sock,"CD-INGEST","OBJECT",rnode);
	}
	//------------------------------------
	// ATRANS
	//------------------------------------
	if (matchlike(vec,"*gave.v*")==1)
	{

		if (matchlike(vec,"* gave.v")==1) addnote(sock,"sent","isa","CD-ATRANS");
		if (matchlike(vec,"* ss gave.v")==1) addnote(sock,"CD-ATRANS","ACTOR",lnode);
		if (matchlike(vec,"gave.v os *.n")==1) addnote(sock,"CD-ATRANS","OBJECT",rnode);
		if (matchlike(vec,"gave.v os* *.n")==1) addnote(sock,"CD-ATRANS","OBJECT",rnode);
	}

	if (matchlike(vec,"*got.v*")==1)
	{
		if (matchlike(vec,"* got.v")==1) addnote(sock,"sent","isa","CD-ATRANS");
		if (matchlike(vec,"* ss got.v")==1) addnote(sock,"CD-ATRANS","ACTOR",lnode);
		if (matchlike(vec,"got.v os *.n")==1) addnote(sock,"CD-ATRANS","OBJECT",rnode);
		if (matchlike(vec,"got.v os* *.n")==1) addnote(sock,"CD-ATRANS","OBJECT",rnode);
	}


	//------------------------------------
	// PTRANS
	//------------------------------------
	if (matchlike(vec,"*went.v*")==1)
	{

		if (matchlike(vec,"* went.v")==1) addnote(sock,"sent","isa","CD-PTRANS");
		if (matchlike(vec,"* ss went.v")==1) addnote(sock,"CD-PTRANS","ACTOR",lnode);
		if (matchlike(vec,"* ss went.v")==1) addnote(sock,"CD-PTRANS","OBJECT",lnode);
		if (matchlike(vec,"went.v k *")==1) addnote(sock,"CD-PTRANS","TO",rnode);
	}
	//------------------------------------
	// POSS
	//------------------------------------
	if (matchlike(vec,"*had.v*")==1)
	{

		if (matchlike(vec,"* had.v")==1) addnote(sock,"sent","isa","CD-POSS");
		if (matchlike(vec,"* ss had.v")==1) addnote(sock,"CD-POSS","ACTOR",lnode);
		if (matchlike(vec,"had.v os *")==1) addnote(sock,"CD-POSS","OBJECT",rnode);
	}

	if (matchlike(vec,"*has.v*")==1)
	{

		if (matchlike(vec,"* has.v")==1) addnote(sock,"sent","isa","CD-POSS");
		if (matchlike(vec,"* ss has.v")==1) addnote(sock,"CD-POSS","ACTOR",lnode);
		if (matchlike(vec,"has.v os *")==1) addnote(sock,"CD-POSS","OBJECT",rnode);
	}
	//------------------------------------
	// ATTRIBUTE
	//------------------------------------
	if (matchlike(vec,"*.a a *.n")==1) addnote(sock,rnode,"ATTRIB",lnode);
	if (matchlike(vec,"*.n an *.n")==1) addnote(sock,rnode,"ATTRIB-CLASS",lnode);

}

void twostepCD(SOCKET sock,char *vec,char *lnode,char *cnode,char *rnode,int LN,int CN, int RN)
{
   char *lxdata;
   char *rxdata;
   char *sp;
   int learnflg;
   unsigned char hashc;
	//-------------------------------------------------------
	// General purpose processor
	rrule_node *rp;
	//OutputDebugString("2stepCD[");
	//OutputDebugString((LPCTSTR) vec);
	//OutputDebugString("]\n");

	learnflg=1;
	hashc=*vec;
	hashc=tolower(hashc);
		// general test
		rp=rruleroot2.ruleroots[hashc];
		while (rp != NULL)
		{
			if (matchlike(vec,rp->pattern)==1) 
			{
				sock_template(sock,rp->templ,lnode,cnode,rnode,LN,CN,RN);
				learnflg=0;
			}
			rp=rp->next;
		}
		// wildcard "?" test
		hashc='?';
		rp=rruleroot2.ruleroots[hashc];
		while (rp != NULL)
		{
			if (matchlike(vec,rp->pattern)==1) 
			{
				sock_template(sock,rp->templ,lnode,cnode,rnode,LN,CN,RN);
				learnflg=0;
			}
			rp=rp->next;
		}
		// wildcard "*" test
		hashc='*';
		rp=rruleroot2.ruleroots[hashc];
		while (rp != NULL)
		{
			if (matchlike(vec,rp->pattern)==1) 
			{
				sock_template(sock,rp->templ,lnode,cnode,rnode,LN,CN,RN);
				learnflg=0;
			}
			rp=rp->next;
		}

	if ((learnflg==1)&&(cyclishQ==1)) guessrule("cycR2.txt",vec,lnode,rnode);

	// Collect trigrams
	if (matchlike(vec,"* m* * j *")==1) 
	{

	sprintf(message,"<trigram>%s %s %s</trigram>\r\n",lnode,cnode,rnode);	
	sp = strstr(kb_trigrams,message);
	if (sp==NULL)
		{
			sprintf(kb_tempb,"%s %s",kb_trigrams,message);
			strcpy(kb_trigrams,kb_tempb);
		}
	}
	return;
	//-------------------------------------------------------

	lxdata=bin_search(lnode,extrakbfile);
	rxdata=bin_search(rnode,extrakbfile);

	//ATRANS
    if (matchlike(vec, "gave.v mvp to j* *" )==1) addnote(sock, "CD-ATRANS","TO", rnode);
	//PTRANS
    if (matchlike(vec, "went.v mvp to j* *" )==1) addnote(sock, "CD-PTRANS","TO", rnode);

}

void singlestep(SOCKET sock,char *vec,char *lnode,char *rnode)
{
   if (outputMode == 4) return; // if CDXML
	//return;

	if (matchlike(vec,"*.a a *")==1) addnote(sock,rnode,"ATTRIBUTE",lnode);
	if (matchlike(vec,"*.v a *.n")==1) addnote(sock,lnode,"ATTRIBUTE",rnode);
	if (matchlike(vec,"*.v af *.n")==1) addnote(sock,lnode,"ATTRIBUTE",rnode);
	if (matchlike(vec,"* an *")==1) addnote(sock,rnode,"ATTRIBUTE",lnode);
	if (matchlike(vec,"*.n al *.n")==1) addnote(sock,rnode,"DETERMINER_ATTRIBUTE",lnode);

	if (matchlike(vec,"*.n b *.n")==1) addnote(sock,rnode,"THEME",lnode);
	if (matchlike(vec,"*.n bi *.n")==1) addnote(sock,rnode,"BEINGQ",lnode);
	if (matchlike(vec,"*.n bt *.n")==1) addnote(sock,rnode,"TIME",lnode);
	if (matchlike(vec,"* bs *.v")==1) addnote(sock,rnode,"THEME",lnode);


	if (matchlike(vec,"* c *.n")==1) addnote(sock,rnode,"SUBJECT",lnode);

	if (matchlike(vec,"* d* *.n")==1) addnote(sock,rnode,"DETERMINER",lnode);
	if (matchlike(vec,"* dg *.n")==1) addnote(sock,rnode,"NAME",lnode);
	if (matchlike(vec,"* dt *.n")==1) addnote(sock,rnode,"TIME",lnode);
	if (matchlike(vec,"* dmcn *.n")==1) addnote(sock,rnode,"QUANITY",lnode);

	if (matchlike(vec,"* em *.v")==1) addnote(sock,lnode,"MANNER",rnode);
	if (matchlike(vec,"*.e e *.v")==1) addnote(sock,rnode,"MANNER",lnode);
	if (matchlike(vec,"*.v e *.e")==1) addnote(sock,lnode,"MANNER",rnode);
	if (matchlike(vec,"*.e ea *")==1) addnote(sock,lnode,"ATTRIBUTE",rnode);

	if (matchlike(vec,"* fm *")==1) addnote(sock,rnode,"SOURCE",lnode);

	if (matchlike(vec,"* g *")==1) addnote(sock,rnode,"ATTRIBUTE",lnode);
	if (matchlike(vec,"* gn *")==1) addnote(sock,rnode,"INSTANCE",lnode);

	if (matchlike(vec,"*.v i* *.v")==1) addnote(sock,rnode,"INFINITIVE",lnode);
	if (matchlike(vec,"* in *")==1) addnote(sock,rnode,"TIME",lnode);

	if (matchlike(vec,"* jq *")==1) addnote(sock,rnode,"QUERY",lnode);
	if (matchlike(vec,"* jt *")==1) addnote(sock,rnode,"TIME",lnode);

	if (matchlike(vec,"*.v k *")==1) addnote(sock,lnode,"PARTICLE",rnode);

	if (matchlike(vec,"* l *")==1) addnote(sock,rnode,"SUPURLATIVE",lnode);
	//if (matchlike(vec,"* li *")==1) addnote(sock,lnode,"ANALOGY",rnode);

	if (matchlike(vec,"* n *")==1) addnote(sock,lnode,"NEGATION",rnode);
	if (matchlike(vec,"* nd *")==1) addnote(sock,rnode,"QUANITY",lnode);
	if (matchlike(vec,"* nr *")==1) addnote(sock,rnode,"FRACTION",lnode);
	if (matchlike(vec,"* ns *")==1) addnote(sock,rnode,"QUANITY",lnode);
	if (matchlike(vec,"* nt *")==1) addnote(sock,rnode,"NEGATION",lnode);
	if (matchlike(vec,"* nw *")==1) addnote(sock,rnode,"FRACTION",lnode);

	
	
	// modifiers
	if (matchlike(vec,"*.v mva *")==1) addnote(sock,lnode,"MANNER",rnode);
	if (matchlike(vec,"*.v mvpn *")==1) addnote(sock,lnode,"SPEC",rnode);
	if (matchlike(vec,"*.v mvpn *.t")==1) addnote(sock,lnode,"TIME-SPEC",rnode);
	if (matchlike(vec,"*.n mp *.a")==1) addnote(sock,lnode,"ATTRIBUTE",rnode);
	if (matchlike(vec,"* mxs *")==1) addnote(sock,lnode,"MODIFIER",rnode);
	if (matchlike(vec,"* mxp *")==1) addnote(sock,lnode,"MODIFIER",rnode);
	if (matchlike(vec,"* mx *")==1) addnote(sock,lnode,"MODIFIER",rnode);
	//if (matchlike(vec,"*.v mvp *")==1) addnote(sock,lnode,"SPEC",rnode);

	// objects
	if (matchlike(vec,"* o* *.v")==1) addnote(sock,rnode,"OBJECT",lnode);
	if (matchlike(vec,"*.v o* *")==1) addnote(sock,lnode,"OBJECT",rnode);
	if (matchlike(vec,"* od *.v")==1) addnote(sock,rnode,"DISTANCE",lnode);
	if (matchlike(vec,"* on *.v")==1) addnote(sock,rnode,"TIME-SPEC",lnode);
	if (matchlike(vec,"* ot *.v")==1) addnote(sock,rnode,"TIME-SPEC",lnode);
	if (matchlike(vec,"*.v od *")==1) addnote(sock,lnode,"DISTANCE",rnode);
	if (matchlike(vec,"*.v on *")==1) addnote(sock,lnode,"TIME-SPEC",rnode);
	if (matchlike(vec,"*.v ot *")==1) addnote(sock,lnode,"TIME-SPEC",rnode);

	if (matchlike(vec,"*.v pa *")==1) addnote(sock,lnode,"MANNER",rnode);
	if (matchlike(vec,"*.v paf *")==1) addnote(sock,lnode,"ATTRIBUTE",rnode);
	if (matchlike(vec,"* pf *")==1) addnote(sock,lnode,"QUERY",rnode);

	// subjects
	if (matchlike(vec,"* ss *.v")==1) addnote(sock,rnode,"THEME",lnode);
	if (matchlike(vec,"* s *.v")==1) addnote(sock,rnode,"THEME",lnode);
	if (matchlike(vec,"* sp *.v")==1) addnote(sock,rnode,"THEME",lnode);
	if (matchlike(vec,"* spx *.v")==1) addnote(sock,rnode,"THEME",lnode);
	if (matchlike(vec,"*.v si* *")==1) addnote(sock,lnode,"THEME",rnode);
	if (matchlike(vec,"* sp*i *.v")==1) addnote(sock,rnode,"AGENT",lnode);
	if (matchlike(vec,"* ss gave.v")==1) addnote(sock,rnode,"SOURCE",lnode);

	if (matchlike(vec,"* pf *")==1) addnote(sock,lnode,"QUERY",rnode);

	if (matchlike(vec,"* q *")==1) addnote(sock,lnode,"QUERY",rnode);
	if (matchlike(vec,"* qi *")==1) addnote(sock,lnode,"QUERY",rnode);
	
	if (matchlike(vec,"* ta *")==1) addnote(sock,lnode,"TIME-SPEC",rnode);
	if (matchlike(vec,"* td *")==1) addnote(sock,lnode,"TIME-SPEC",rnode);
	if (matchlike(vec,"* ti *")==1) addnote(sock,lnode,"TITLE",rnode);
	if (matchlike(vec,"* tm *")==1) addnote(sock,lnode,"TIME-SPEC",rnode);
	if (matchlike(vec,"* tq *")==1) addnote(sock,lnode,"TIME-DETERMINER",rnode);
	if (matchlike(vec,"* tw *")==1) addnote(sock,lnode,"TIME-SPEC",rnode);
	if (matchlike(vec,"* ty *")==1) addnote(sock,lnode,"TIME-SPEC",rnode);


	if (matchlike(vec,"* un *")==1) addnote(sock,lnode,"TIME-CONDITION",rnode);

	if (matchlike(vec,"* wn *")==1) addnote(sock,lnode,"TIME-WHEN",rnode);
	if (matchlike(vec,"* wr *")==1) addnote(sock,lnode,"LOCATION-QUERY",rnode);

	if (matchlike(vec,"* y *")==1) addnote(sock,lnode,"QUANITY",rnode);
	if (matchlike(vec,"* yd *")==1) addnote(sock,lnode,"DISTANCE",rnode);
	if (matchlike(vec,"* yt *")==1) addnote(sock,lnode,"TIME",rnode);
	if (matchlike(vec,"* ya *")==1) addnote(sock,lnode,"SPATIAL-SPEC",rnode);
	if (matchlike(vec,"* ye *")==1) addnote(sock,lnode,"QUANITY-SPEC",rnode);
//	if (matchlike(vec,"* ys *")==1) addnote(sock,lnode,"TIME-SPEC",rnode);

	//nstep(sock,1,vec,lnode,rnode);
}

void twostep(SOCKET sock,char *vec,char *lnode,char *rnode)
{
   char *lxdata;
   char *rxdata;
   
   if (outputMode == 4) return; // if CDXML

	lxdata=bin_search(lnode,extrakbfile);
	rxdata=bin_search(rnode,extrakbfile);

	//SS+PV
    if (matchlike(vec, "* ss was.v pv *.v" )==1) addnote(sock, rnode, "THEME", lnode);
    if (matchlike(vec, "* ss is.v pgfb *.v" )==1) addnote(sock, rnode, "AGENT", lnode);
    if (matchlike(vec, "* ss was.v opt *.a" )==1) addnote(sock, lnode, "ATTRIBUTE", rnode);
    //SS+MVP (pat ss communicated.v mpv to )
//    if (matchlike(vec, "* ss *.v mvp *" )==1) addnote(sock, Rlink$(p1), "AGENT", lnode);
    //SP+I
//     if (matchlike(vec, "* sp *.v i *.v" )==1) addnote(sock, Rlink$(p1), "AGENT", lnode);
    //MG+Pv
    if (matchlike(vec, "* mg *.v pv *.v" )==1) addnote(sock, rnode, "THEME", lnode);
   //mp + J*
    if (matchlike(vec, "* mp with js *" )==1) addnote(sock, lnode, "PART", rnode);
    if (matchlike(vec, "* mp with jp *" )==1) addnote(sock, lnode, "ATTRIBUTE", rnode);
    if (matchlike(vec, "*.n mp with js *" )==1) addnote(sock, lnode, "SPEC", rnode);

	if (matchlike(vec,"* of *")==1)
	{
		if (matchlike(vec, "* mp of j* *" )==1)
		{
			if (rxdata!=NULL && strstr(rxdata,"location")!=NULL)
					addnote(sock, lnode, "SPEC-AREA", rnode);
			else
				addnote(sock, lnode, "SPEC", rnode);

		}

		if (matchlike(vec, "* mg of jg *" )==1) addnote(sock, lnode, "ORIGIN", rnode);
	}
    //MP + PV
    if (matchlike(vec, "* mp is.v pv *.v" )==1) addnote(sock, rnode, "THEME", lnode);

    //MVP+j*
	
    if (matchlike(vec, "* mvp with j* *" )==1)
	{
		if (isPronoun(rnode)!=0)
			addnote(sock, lnode, "COAGENT", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"person")!=NULL)
				addnote(sock, lnode, "COAGENT", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"social_group")!=NULL)
				addnote(sock, lnode, "COAGENT", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"artiface")!=NULL)
				addnote(sock, lnode, "INSTURMENT", rnode);
		else
			addnote(sock, lnode, "COAGENT-INSTURMENT", rnode);
    //if (matchlike(vec, "* mvp with js *" )==1) addnote(sock, lnode, "COAGENT-INSTURMENT", rnode);
    //if (matchlike(vec, "* mvp with j *" )==1) addnote(sock, lnode, "COAGENT-INSTURMENT", rnode);
	}

    if (matchlike(vec, "* mvp for j* *" )==1)
	{
		if (isPronoun(rnode)!=0)
			addnote(sock, lnode, "BENIFICIARY", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"person")!=NULL)
				addnote(sock, lnode, "BENIFICIARY", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"social_group")!=NULL)
				addnote(sock, lnode, "BENIFICIARY", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"artiface")!=NULL)
				addnote(sock, lnode, "TARGET", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"time")!=NULL)
				addnote(sock, lnode, "DURATION", rnode);
		else
			addnote(sock, lnode, "RESULT", rnode);

	}

    if (matchlike(vec, "* mvp from j* *" )==1)
	{
		if (isPronoun(rnode)!=0)
			addnote(sock, lnode, "SOURCE", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"person")!=NULL)
				addnote(sock, lnode, "SOURCE", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"social_group")!=NULL)
				addnote(sock, lnode, "SOURCE", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"artiface")!=NULL)
				addnote(sock, lnode, "SOURCE", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"time")!=NULL)
				addnote(sock, lnode, "START-TIME", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"location")!=NULL)
				addnote(sock, lnode, "ORIGIN", rnode);
		else
			addnote(sock, lnode, "SOURCE", rnode);

	}
    if (matchlike(vec, "* mvp to j* *" )==1)
	{
		if (isPronoun(rnode)!=0)
			addnote(sock, lnode, "RECIPIENT", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"person")!=NULL)
				addnote(sock, lnode, "RECIPIENT", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"social_group")!=NULL)
				addnote(sock, lnode, "RECIPIENT", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"artiface")!=NULL)
				addnote(sock, lnode, "SOURCE", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"time")!=NULL)
				addnote(sock, lnode, "END-TIME", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"location")!=NULL)
				addnote(sock, lnode, "DESTINATION", rnode);
		else
			addnote(sock, lnode, "DESTINATION", rnode);

    //if (matchlike(vec, "*.v mvp to js *" )==1) addnote(sock, lnode, "DESTINATION", rnode);
    //if (matchlike(vec, "* mvp to jp *" )==1) addnote(sock, lnode, "DESTINATION", rnode);
    //if (matchlike(vec, "* mvp to j *" )==1) addnote(sock, lnode, "RECIPIENT", rnode);
	}
    if (matchlike(vec, "* mvp by j* *" )==1)
	{
		if (isPronoun(rnode)!=0)
			addnote(sock, lnode, "AGENT", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"person")!=NULL)
				addnote(sock, lnode, "AGENT", rnode);
		else
			if (rxdata!=NULL && strstr(rxdata,"social_group")!=NULL)
				addnote(sock, lnode, "AGENT", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"artiface")!=NULL)
				addnote(sock, lnode, "METHOD", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"time")!=NULL)
				addnote(sock, lnode, "END-TIME", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"location")!=NULL)
				addnote(sock, lnode, "PATH", rnode);
		else
			addnote(sock, lnode, "METHOD", rnode);

    //if (matchlike(vec, "* mvp by js *" )==1) addnote(sock, lnode, "METHOD", rnode);
	}

    if (matchlike(vec, "*.v li like* j* *" )==1) addnote(sock, lnode, "ANALOGY", rnode);

    if (matchlike(vec, "* mvp at* j* *" )==1) addnote(sock, lnode, "LOCATION", rnode);
    if (matchlike(vec, "* mvp against* j* *" )==1) addnote(sock, lnode, "CONTACT", rnode);
    if (matchlike(vec, "* mvp about* js *" )==1) addnote(sock, lnode, "TOPIC", rnode);
    if (matchlike(vec, "*.v mvp on js *" )==1) addnote(sock, lnode, "TARGET-LOCATION-AREA", rnode);

    // very broad rule "talk about implementing", "cooperate on reducing"
	if (matchlike(vec, "*.v mvp * mgp *.v" )==1) addnote(sock, lnode, "GOAL-ACTION", rnode);

	// IN
 	if (matchlike(vec, "* mvp in * *" )==1)
	{
	if (rxdata!=NULL && strstr(rxdata,"location")!=NULL)
				addnote(sock, lnode, "LOCATION", rnode);
	else
		if (rxdata!=NULL && strstr(rxdata,"time")!=NULL)
				addnote(sock, lnode, "TIMEPERIOD", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"artiface")!=NULL)
				addnote(sock, lnode, "MEDIUM", rnode);
		else 
			if (rxdata!=NULL && strstr(rxdata,"physical_object")!=NULL)
				addnote(sock, lnode, "MEDIUM", rnode);

		addnote(sock, lnode, "LOCATION-TIME", rnode);
	//if (matchlike(vec, "* mvp in jp *" )==1) addnote(sock, lnode, "LOCATION-TIME", rnode);
    //if (matchlike(vec, "* mvp in js *" )==1) addnote(sock, lnode, "MEDIUM", rnode);
	}

	if (matchlike(vec, "* mvp inside* j* *" )==1) addnote(sock, lnode, "LOCATION", rnode);
	if (matchlike(vec, "* mvp into* j* *" )==1) addnote(sock, lnode, "LOCATION", rnode);
   //M
    if (matchlike(vec, "is.v pp for.p j *" )==1) addnote(sock, lnode, "TARGET-BENIFICIARY", rnode);
	// to
	if (matchlike(vec,"*.v to to i *.v")==1) addnote(sock,rnode,"INFINITIVE",lnode);

	// "The rose is red." = "rose ss is.v pa red.a"
	if (matchlike(vec,"* ss is.v pa *.a")==1) addnote(sock,lnode,"ATTRIBUTE",rnode);


	
	// CEMA like notes
	// possesive "jane's car"
	if (matchlike(vec,"* y* 's.p d* *")==1) addnote3(sock,lnode,rnode,"ROWN");
	if (matchlike(vec,"* s* 's.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
	
	// general case for subj-verb-object
	if (matchlike(vec,"* s* *.v o* *")==1)
	{
		if (matchlike(vec,"* s* gave.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* got.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* get.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* sent.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* sending.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* bought.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* buy.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* sold.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* provid*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* has.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* have.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* acquir*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* need*.v o* *")==1) addnote3(sock,lnode,rnode,"RNEED");
		if (matchlike(vec,"* s* requir*.v o* *")==1) addnote3(sock,lnode,rnode,"RNEED");
		if (matchlike(vec,"* s* required.v o* *")==1) addnote3(sock,lnode,rnode,"RNEED");
		if (matchlike(vec,"* s* supply.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");

		if (matchlike(vec,"* s* forward*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* dispatch*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* dispatch*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* ship*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* transport*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");

		if (matchlike(vec,"* s* provid*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* suppl*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* took.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* take.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* taking.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");

		if (matchlike(vec,"* s* steal*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* stole.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* feed*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* offer*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* fuel*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* refuel*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* equip*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* outfit*.v o* *")==1) addnote3(sock,lnode,rnode,"ROUT");
		if (matchlike(vec,"* s* want*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* demand*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* requsition*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* capture*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* catch*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* caught*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* hunt*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* trap*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* entrap*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* ensnare*.v o* *")==1) addnote3(sock,lnode,rnode,"RUNKW");
		if (matchlike(vec,"* s* enslave*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* lease*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* rent*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* hire*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* receive*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* find*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* found*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* accept*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* adopt*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* obtain*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* excavat*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* procur*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* secur*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* earn*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* collect*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* pilfer*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* shoplift*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* pirat*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* embezzel*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* misappropriat*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* rob*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* loot*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* plunder*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* burgl*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* heist*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");
		if (matchlike(vec,"* s* own*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* exhibit*.v o* *")==1) addnote3(sock,lnode,rnode,"ROWN");
		if (matchlike(vec,"* s* loot*.v o* *")==1) addnote3(sock,lnode,rnode,"RIN");


	}
	//nstep(sock,2,vec,lnode,rnode);
}

int findrecipient(Linkage linkage,int cur_vb,int n_links)
{
	int l;
	int rwp;
	int ans;
	char *klabel;
	char *rword;
	// chain through give-mvp-to-js-mary
	// process is a recursive search of the linkage tree
	// so multiple paths could be explored
	for (l=0;l<n_links;l++)
	{
		if (linkage_get_link_lword(linkage, l)==cur_vb)
		{
		klabel  = linkage_get_link_label(linkage, l);
		rwp  = linkage_get_link_rword(linkage, l);

		if ((strcmpi(klabel,"js")==0)
			||(strcmpi(klabel,"j")==0)
			||(strcmpi(klabel,"jp")==0)
			||(strcmpi(klabel,"jg")==0)
			)
			return (rwp); // base case: we found it

		rword=linkage_get_word(linkage,rwp);
		if  ((strcmpi(klabel,"mvp")==0)&&(strcmpi(rword,"to")==0)) 
			{
				// normally a recursive call
				// use a recursive call here so its a search process
				ans= findrecipient(linkage,rwp,n_links);
				if (ans!=0) return ans;
			}

		}
	}
	return (0); // never found it or too complex for us now

}

int findsource(Linkage linkage,int cur_vb,int n_links)
{
	int l;
	int rwp;
	int ans;
	char *klabel;
	char *rword;
	// chain through give-mvp-to-js-mary
	// process is a recursive search of the linkage tree
	// so multiple paths could be explored
	for (l=0;l<n_links;l++)
	{
		if (linkage_get_link_lword(linkage, l)==cur_vb)
		{
		klabel  = linkage_get_link_label(linkage, l);
		rwp  = linkage_get_link_rword(linkage, l);

		if ((strcmpi(klabel,"js")==0)
			||(strcmpi(klabel,"j")==0)
			||(strcmpi(klabel,"jp")==0)
			||(strcmpi(klabel,"jg")==0)
			)
			return (rwp); // base case: we found it

		rword=linkage_get_word(linkage,rwp);
		if  ((strcmpi(klabel,"mvp")==0)&&(strcmpi(rword,"from")==0)) 
			{
				// normally a recursive call
				// use a recursive call here so its a search process
				ans= findsource(linkage,rwp,n_links);
				if (ans!=0) return ans;
			}

		}
	}
	return (0); // never found it or too complex for us now

}

int rootverb(Linkage linkage,int cur_vb,int n_links)
{
	int vb;
	int l;
	char *klabel;
	// chain through John=(ss)-will=(if)-have=(to)-to-(i)-give
	for (l=0;l<n_links;l++)
	{
		if (linkage_get_link_lword(linkage, l)==cur_vb)
		{
		klabel  = linkage_get_link_label(linkage, l);
		if ((strcmpi(klabel,"if")==0) 
			||(strcmpi(klabel,"to")==0)
			||(strcmpi(klabel,"too")==0)
			||(strcmpi(klabel,"i")==0)
			||(strcmpi(klabel,"i*d")==0)
			||(strcmpi(klabel,"ix")==0)
			||(strcmpi(klabel,"pv")==0)
			||(strcmpi(klabel,"pg")==0)
			||(strcmpi(klabel,"pa")==0)
			||(strcmpi(klabel,"pvf")==0)			
			||(strcmpi(klabel,"pg*b")==0)
			||(strcmpi(klabel,"pp")==0))
			{
				vb= linkage_get_link_rword(linkage, l);
				// normally a recursive call
				// return rootverb(linkage,vb,n_links); 
				// for effiency reset loop with new verb
				l=-1; // post loop increment will bump up to zero
				cur_vb=vb;
			}
		}
	}
	return (cur_vb);
}


int phrasechain(Linkage linkage,int cur_vb,int n_links)
{
	int vb;
	int l;
	char *klabel;
	// chain through John=(ss)-will=(if)-have=(to)-to-(i)-give
	for (l=0;l<n_links;l++)
	{
		if (linkage_get_link_rword(linkage, l)==cur_vb)
		{
		klabel  = linkage_get_link_label(linkage, l);
		if ((strcmpi(klabel,"an")==0) 
			||(strcmpi(klabel,"g")==0)
			||(strcmpi(klabel,"en")==0)
			||(strcmpi(klabel,"nn")==0)
			||(strcmpi(klabel,"ta")==0)
			||(strcmpi(klabel,"ty")==0)
			||(strcmpi(klabel,"if")==0)
			||(strcmpi(klabel,"i")==0)
			||(strcmpi(klabel,"ix")==0)
			||(strcmpi(klabel,"pgf")==0)			
			||(strcmpi(klabel,"pvf")==0)
			||(strcmpi(klabel,"pv")==0)
			||(strcmpi(klabel,"ppf")==0)
			||(strcmpi(klabel,"idxd")==0)
			||(strcmpi(klabel,"dtn")==0))
			{
				vb= linkage_get_link_lword(linkage, l);
				// normally a recursive call
				// return rootverb(linkage,vb,n_links); 
				// for effiency reset loop with new verb
				l=-1; // post loop increment will bump up to zero
				cur_vb=vb;
			}
		}
	}
	return (cur_vb);
}


void inferlinks_common(SOCKET sock, Linkage linkage)
{
  int N_links;
  int link,link2,link3;
  int l,r,l2,r2,l3,r3;
  char *klabel, *llabel, *rlabel;
  char *klabel2, *llabel2, *rlabel2;
  char *klabel3, *llabel3, *rlabel3;
  char *lnode;
  char *rnode;
  int base_verb;
  int recipient;
  int source;
	
    kb_buf[0]=0;
	kb_tempb[0]=0;

   N_links = linkage_get_num_links(linkage);
	linkcnt=0;
	if ((cyclishQ==1) && (outputMode !=4))
	{
		for(link=0;link<N_links;link++)
		{
			l      = linkage_get_link_lword(linkage, link);
			r      = linkage_get_link_rword(linkage, link);
			klabel  = linkage_get_link_label(linkage, link);
			llabel = linkage_get_link_llabel(linkage, link);
			rlabel = linkage_get_link_rlabel(linkage, link);
			//addnote(sock,llabel,klabel,rlabel);
			lnode =linkage_get_word(linkage,l);
			sprintf(simplevec,"%s %s %s",lnode,klabel,linkage_get_word(linkage,r));
			toLower(simplevec);
			addnote(sock,lnode,klabel,linkage_get_word(linkage,r));
		}

	}

	for(link=0;link<N_links;link++)
	{
		l      = linkage_get_link_lword(linkage, link);
		r      = linkage_get_link_rword(linkage, link);
		klabel  = linkage_get_link_label(linkage, link);
		llabel = linkage_get_link_llabel(linkage, link);
		rlabel = linkage_get_link_rlabel(linkage, link);
		//addnote(sock,llabel,klabel,rlabel);
		lnode =linkage_get_word(linkage,l);
		sprintf(simplevec,"%s %s %s",lnode,klabel,linkage_get_word(linkage,r));
		toLower(simplevec);
		//sock_print(sock,"<svec>");
		//sock_print(sock,simplevec);
		//sock_print(sock,"</svec>\r\n");
		if (cyclishQ==1)
		{
			//addnote(sock,lnode,klabel,linkage_get_word(linkage,r));
		}

		singlestep(sock,simplevec,linkage_get_word(linkage,l),linkage_get_word(linkage,r));
		singlestepCD(sock,simplevec,linkage_get_word(linkage,l),linkage_get_word(linkage,r),l,r);
		
		for(link2=0;link2<N_links;link2++)
		{
			l2      = linkage_get_link_lword(linkage, link2);
			if ((link2!=link)&&
				(l2==r) )
			{
				// second stage linkage
				r2      = linkage_get_link_rword(linkage, link2);
				klabel2  = linkage_get_link_label(linkage, link2);
				llabel2 = linkage_get_link_llabel(linkage, link2);
				rlabel2 = linkage_get_link_rlabel(linkage, link2);
				rnode=linkage_get_word(linkage,r2);
				sprintf(pairvec,"%s %s %s",simplevec,llabel2,rnode);
				toLower(pairvec);
				//sock_print(sock,"<pvec>");
				//sock_print(sock,pairvec);
				//sock_print(sock,"</pvec>\r\n");
				
				twostepCD(sock,pairvec,lnode,linkage_get_word(linkage,r),rnode,l,r,r2);
				twostep(sock,pairvec,lnode,rnode);
				//*.n ss was.v Pg*b getting.v (PV *.v) => *.v Theme *.
				if (matchlike(pairvec,"* s* *.v p* *.v")==1)
				{
					for(link3=0;link3<N_links;link3++)
					{
						l3      = linkage_get_link_lword(linkage, link3);
						if ((link3!=link)&&(link3!=link2)&&
							(l3==r2))
						{
							klabel3=linkage_get_link_label(linkage, link3);
							if (matchlike(klabel3,"p*")==1)
							{
								r3= linkage_get_link_rword(linkage, link3);
								addnote(sock,linkage_get_word(linkage,r3),"THEME",linkage_get_word(linkage,l));
							}
						}
					}

				}
				//*.n sp need.v to to (I *.v) => *.v Agent *.n

				if (matchlike(pairvec,"* s* *.v to *")==1)
				{
					for(link3=0;link3<N_links;link3++)
					{
						l3      = linkage_get_link_lword(linkage, link3);
						if ((link3!=link)&&(link3!=link2)&&
							(l3==r2))
						{
							klabel3=linkage_get_link_label(linkage, link3);
							if (matchlike(klabel3,"p*")==1)
							{
								r3= linkage_get_link_rword(linkage, link3);
								addnote(sock,linkage_get_word(linkage,r3),"AGENT",linkage_get_word(linkage,l));
							}
						}
					}

				}


			
			}
		}

	
	// check modal chains 
		//"john ss should.v I give.v" -- normal subject
		if ((matchlike(simplevec,"* ss *.v")==1)
			||(matchlike(simplevec,"* s *.v")==1)
			||(matchlike(simplevec,"* sp *.v")==1)
			||(matchlike(simplevec,"* spx *.v")==1)
			||(matchlike(simplevec,"* bs *.v")==1)
			||(matchlike(simplevec,"* sp*i *.v")==1))
		{
			base_verb=rootverb(linkage,r,N_links);
			if (base_verb!=r)
			{ // there is a verb beyond ours
				sprintf(simplevec,"%s %s %s",lnode,klabel,linkage_get_word(linkage,base_verb));
				toLower(simplevec);
				singlestep(sock,simplevec,linkage_get_word(linkage,l),linkage_get_word(linkage,base_verb));
				singlestepCD(sock,simplevec,linkage_get_word(linkage,l),linkage_get_word(linkage,base_verb),l,base_verb);
				 // second level linking
					for(link2=0;link2<N_links;link2++)
					{
						l2      = linkage_get_link_lword(linkage, link2);
						if ((link2!=link)&&
							(l2==base_verb) )
						{
							// second stage linkage
							r2      = linkage_get_link_rword(linkage, link2);
							klabel2  = linkage_get_link_label(linkage, link2);
							llabel2 = linkage_get_link_llabel(linkage, link2);
							rlabel2 = linkage_get_link_rlabel(linkage, link2);
							rnode=linkage_get_word(linkage,r2);
							sprintf(pairvec,"%s %s %s",simplevec,llabel2,rnode);
							toLower(pairvec);
							//sock_print(sock,"<pvec>");
							//sock_print(sock,pairvec);
							//sock_print(sock,"</pvec>\r\n");
							
							twostep(sock,pairvec,lnode,rnode);
							twostepCD(sock,pairvec,lnode,linkage_get_word(linkage,base_verb),rnode,l,base_verb,r2);
						}
					}
			
			}
		}
		//"what should sis john give.v" -- inverted subject
		if ((matchlike(simplevec,"*.v sis *")==1)
			||(matchlike(simplevec,"*.v sip*i *")==1)
			||(matchlike(simplevec,"*.v sip *")==1)
			)
		{
			base_verb=rootverb(linkage,l,N_links);
			if (base_verb!=l)
			{ // there is a verb beyond ours
				sprintf(simplevec,"%s ss %s",linkage_get_word(linkage,r),linkage_get_word(linkage,base_verb));
				toLower(simplevec);
				singlestep(sock,simplevec,linkage_get_word(linkage,r),linkage_get_word(linkage,base_verb));
				singlestepCD(sock,simplevec,linkage_get_word(linkage,r),linkage_get_word(linkage,base_verb),r,base_verb);
				 // second level linking
					for(link2=0;link2<N_links;link2++)
					{
						l2      = linkage_get_link_lword(linkage, link2);
						if ((link2!=link)&&
							(l2==base_verb) )
						{
							// second stage linkage
							r2      = linkage_get_link_rword(linkage, link2);
							klabel2  = linkage_get_link_label(linkage, link2);
							llabel2 = linkage_get_link_llabel(linkage, link2);
							rlabel2 = linkage_get_link_rlabel(linkage, link2);
							rnode=linkage_get_word(linkage,r2);
							sprintf(pairvec,"%s %s %s",simplevec,llabel2,rnode);
							toLower(pairvec);
							//sock_print(sock,"<pvec>");
							//sock_print(sock,pairvec);
							//sock_print(sock,"</pvec>\r\n");
							
							twostep(sock,pairvec,linkage_get_word(linkage,r),rnode);
							twostepCD(sock,pairvec,linkage_get_word(linkage,r),linkage_get_word(linkage,base_verb),rnode,r,base_verb,r2);
						}
					}
			}
		}

		//CEMA notes: recipients have positive balance
		// gave give
		// provide provided provides
		// supply render furnish
		//feed
		// offer
		// fuel refuel
		// equip, outfit
		// need, want, require
		// demand, call for, take,exact,
		// requsition
		// acquire, capture,catch
		// hunt,trap, entrap, ensare
		// lease,rent,hire, receive,have,accept
		// buy
		// find
		// accept,adopt, obtain
		// excavate, procure, secure
		// earn
		// collect
		// get
		// steal, stole, pilfer, swipe, shiplift
		// pirate, embezzel, misappropriate
		// rob
		// loot, plunder
		// burglarize, burgle, heist
		// own, have, posses, exhibit
		// sent, forward, dispatch, ship, transport



		if (matchlike(simplevec,"*.v o* *")==1)
		{
			if ((matchlike(simplevec,"give.v o* *")==1)
				||(matchlike(simplevec,"gave.v o* *")==1)

				||(matchlike(simplevec,"sent.v o* *")==1)
				||(matchlike(simplevec,"send*.v o* *")==1)
				||(matchlike(simplevec,"forward*.v o* *")==1)
				||(matchlike(simplevec,"dispatch*.v o* *")==1)
				||(matchlike(simplevec,"dispatch*.v o* *")==1)
				||(matchlike(simplevec,"ship*.v o* *")==1)
				||(matchlike(simplevec,"transport*.v o* *")==1)

				||(matchlike(simplevec,"sold.v o* *")==1)
				||(matchlike(simplevec,"provid*.v o* *")==1)
				||(matchlike(simplevec,"acquir*.v o* *")==1)
				||(matchlike(simplevec,"need*.v o* *")==1)
				||(matchlike(simplevec,"requir*.v o* *")==1)
				||(matchlike(simplevec,"suppl*.v o* *")==1)
				||(matchlike(simplevec,"get.v o* *")==1)
				||(matchlike(simplevec,"got.v o* *")==1)
				||(matchlike(simplevec,"took.v o* *")==1)
				||(matchlike(simplevec,"take.v o* *")==1)
				||(matchlike(simplevec,"taking.v o* *")==1)
				||(matchlike(simplevec,"buy.v o* *")==1)
				||(matchlike(simplevec,"bought.v o* *")==1)

				||(matchlike(simplevec,"steal*.v o* *")==1)
				||(matchlike(simplevec,"stole.v o* *")==1)
				||(matchlike(simplevec,"feed*.v o* *")==1)
				||(matchlike(simplevec,"offer*.v o* *")==1)
				||(matchlike(simplevec,"fuel*.v o* *")==1)
				||(matchlike(simplevec,"refuel*.v o* *")==1)
				||(matchlike(simplevec,"equip*.v o* *")==1)
				||(matchlike(simplevec,"outfit*.v o* *")==1)
				||(matchlike(simplevec,"want*.v o* *")==1)
				||(matchlike(simplevec,"demand*.v o* *")==1)
				||(matchlike(simplevec,"requsition*.v o* *")==1)
				||(matchlike(simplevec,"capture*.v o* *")==1)
				||(matchlike(simplevec,"catch*.v o* *")==1)
				||(matchlike(simplevec,"caught*.v o* *")==1)
				||(matchlike(simplevec,"hunt*.v o* *")==1)
				||(matchlike(simplevec,"trap*.v o* *")==1)
				||(matchlike(simplevec,"entrap*.v o* *")==1)
				||(matchlike(simplevec,"ensnare*.v o* *")==1)
				||(matchlike(simplevec,"enslave*.v o* *")==1)
				||(matchlike(simplevec,"lease*.v o* *")==1)
				||(matchlike(simplevec,"rent*.v o* *")==1)
				||(matchlike(simplevec,"hire*.v o* *")==1)
				||(matchlike(simplevec,"receive*.v o* *")==1)
				||(matchlike(simplevec,"find*.v o* *")==1)
				||(matchlike(simplevec,"found*.v o* *")==1)
				||(matchlike(simplevec,"accept*.v o* *")==1)
				||(matchlike(simplevec,"adopt*.v o* *")==1)
				||(matchlike(simplevec,"obtain*.v o* *")==1)
				||(matchlike(simplevec,"excavat*.v o* *")==1)
				||(matchlike(simplevec,"procur*.v o* *")==1)
				||(matchlike(simplevec,"secur*.v o* *")==1)
				||(matchlike(simplevec,"earn*.v o* *")==1)
				||(matchlike(simplevec,"collect*.v o* *")==1)
				||(matchlike(simplevec,"pilfer*.v o* *")==1)
				||(matchlike(simplevec,"shoplift*.v o* *")==1)
				||(matchlike(simplevec,"pirat*.v o* *")==1)
				||(matchlike(simplevec,"embezzel*.v o* *")==1)
				||(matchlike(simplevec,"misappropriat*.v o* *")==1)
				||(matchlike(simplevec,"rob*.v o* *")==1)
				||(matchlike(simplevec,"loot*.v o* *")==1)
				||(matchlike(simplevec,"plunder*.v o* *")==1)
				||(matchlike(simplevec,"burgl*.v o* *")==1)
				||(matchlike(simplevec,"heist*.v o* *")==1)
				||(matchlike(simplevec,"own*.v o* *")==1)
				||(matchlike(simplevec,"exhibit*.v o* *")==1)
				||(matchlike(simplevec,"loot*.v o* *")==1)
				)
			{
				source=findsource(linkage,l,N_links);
				if (source!=0)
				{ // there is a verb beyond ours
					addnote3(sock,linkage_get_word(linkage,source),linkage_get_word(linkage,r),"ROUT");
				}

				recipient=findrecipient(linkage,l,N_links);
				if (recipient!=0)
				{ // there is a verb beyond ours
					addnote3(sock,linkage_get_word(linkage,recipient),linkage_get_word(linkage,r),"RIN");
				}
			}
		}

	}

	if (outputMode!=1)
	{
		if(cyclishQ==0)
		{
		sprintf(message,"<linkcount>%i</linkcount>\r\n",linkcnt);
		sock_print(sock,message);
		}
	}

}


void nstep(SOCKET sock,int len, char *vec,char *lnode,char *rnode)
{
	//return;
// automatically generated rules
if (len==1)
{
/*
 if((len==1)&& (matchlike(vec,"'s.v object *.n" )==1)) addnote2(sock,"becoming-aware","ground",rnode);
 if((len==1)&& (matchlike(vec,"'s.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"'s.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"'s.v object *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"'s.v object *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"'s.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"'s.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"'s.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"'s.v theme *.n" )==1)) addnote2(sock,"commitment","addressee",rnode);
 if((len==1)&& (matchlike(vec,"'s.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"'s.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"'tis[?].v object *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 */
 if((len==1)&& (matchlike(vec,"*.n attribute dying.v" )==1)) addnote2(sock,"experiencer-obj","cause",lnode);
 if((len==1)&& (matchlike(vec,"*.n attribute dying.v" )==1)) addnote2(sock,"judgment","evaluee",lnode);
 if((len==1)&& (matchlike(vec,"*.n attribute dying.v" )==1)) addnote2(sock,"perception_noise","cause",lnode);
 if((len==1)&& (matchlike(vec,"*.n attribute soaring.v" )==1)) addnote2(sock,"request","message",lnode);
 if((len==1)&& (matchlike(vec,"abduct.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"abduct.v theme *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"abetted.v theme *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"abhor.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"abhorred.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"abhorred.v medium *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"accompany.v location *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"ache.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"ached.v location-time *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"ached.v source *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"ached.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"aching.v attribute *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"aching.v attribute *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"aching.v coagent-insturment *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"addressed.v recipient *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"addressing.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"admire.v destination *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"admire.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"admired.v theme *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"adopted.v theme *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"adored.v attribute *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"adored.v coagent-insturment *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"adored.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"adored.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"advise.v coagent *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"advise.v location-time *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"advise.v object *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"affected.v agent *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"affronted[!].v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"alarmed.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"alarmed.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"alleged.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"alleges.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"alleging.v attribute *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"alleviate.v theme *.n" )==1)) addnote2(sock,"cure","treatment",rnode);
 if((len==1)&& (matchlike(vec,"alleviating.v object *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"allow.v object *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"allowed.v object *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"amazes.v object *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"analysed.v agent *.n" )==1)) addnote2(sock,"scrutiny","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"analysed.v theme *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"anger.v theme *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"anger.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"angled.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"annihilates.v theme *.n" )==1)) addnote2(sock,"death","agent",rnode);
 if((len==1)&& (matchlike(vec,"announcing.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"annoyed.v coagent *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"anointing.v coagent-insturment *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"anointing.v object *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"answer.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"answer.v object *.n" )==1)) addnote2(sock,"communication_response","trigger",rnode);
 if((len==1)&& (matchlike(vec,"answered.v coagent-insturment *.n" )==1)) addnote2(sock,"communication_response","message",rnode);
 if((len==1)&& (matchlike(vec,"answered.v object *.n" )==1)) addnote2(sock,"communication_response","medium",rnode);
 if((len==1)&& (matchlike(vec,"antagonize.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"antagonize.v object *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"anticipate.v theme *.n" )==1)) addnote2(sock,"expectation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"anticipated.v object *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"anticipating.v object *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"appeared.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"appeared.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"appearing.v destination *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"applaud.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"applauded.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"applauded.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"applauded.v theme *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"appointed.v method *.n" )==1)) addnote2(sock,"change-of-leadership","selector",rnode);
 if((len==1)&& (matchlike(vec,"appointed.v object *.n" )==1)) addnote2(sock,"change-of-leadership","new leader",rnode);
 if((len==1)&& (matchlike(vec,"appointed.v theme *.n" )==1)) addnote2(sock,"change-of-leadership","new leader",rnode);
 if((len==1)&& (matchlike(vec,"appoints.v object *.n" )==1)) addnote2(sock,"change-of-leadership","new leader",rnode);
 if((len==1)&& (matchlike(vec,"appoints.v theme *.n" )==1)) addnote2(sock,"change-of-leadership","selector",rnode);
 if((len==1)&& (matchlike(vec,"appreciate.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"approach.v object *.n" )==1)) addnote2(sock,"arriving","goal",rnode);
 if((len==1)&& (matchlike(vec,"approach.v object *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"approached.v object *.n" )==1)) addnote2(sock,"motion_noise","goal",rnode);
 if((len==1)&& (matchlike(vec,"approx[?].v object *.n" )==1)) addnote2(sock,"expansion","result size",rnode);
 if((len==1)&& (matchlike(vec,"arch.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 /*
 if((len==1)&& (matchlike(vec,"are.v object *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"are.v object *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"are.v object *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"are.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"cotheme","cotheme",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"duplication","creator",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"replacement","themes",rnode);
 if((len==1)&& (matchlike(vec,"are.v theme *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 */
   if((len==1)&& (matchlike(vec,"argue.v coagent *.n" )==1)) addnote2(sock,"conversation","interlocutor-2",rnode);
 if((len==1)&& (matchlike(vec,"argue.v theme *.n" )==1)) addnote2(sock,"conversation","interlocutors",rnode);
 if((len==1)&& (matchlike(vec,"armed.v attribute *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"arrayed.v location-time *.n" )==1)) addnote2(sock,"arriving","manner",rnode);
 if((len==1)&& (matchlike(vec,"arrayed.v location-time *.n" )==1)) addnote2(sock,"perception-experience","manner",rnode);
 if((len==1)&& (matchlike(vec,"arrive.v destination *.n" )==1)) addnote2(sock,"arriving","goal",rnode);
 if((len==1)&& (matchlike(vec,"arrive.v location *.n" )==1)) addnote2(sock,"arriving","goal",rnode);
 if((len==1)&& (matchlike(vec,"arrived.v source *.n" )==1)) addnote2(sock,"arriving","source",rnode);
 if((len==1)&& (matchlike(vec,"arrived.v theme *.n" )==1)) addnote2(sock,"transportation","driver",rnode);
 if((len==1)&& (matchlike(vec,"ascertain.v object *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"asked.v theme *.n" )==1)) addnote2(sock,"hear","speaker",rnode);
 if((len==1)&& (matchlike(vec,"asked.v theme *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"asking.v object *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"asking.v object *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"aspire.v destination *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"aspired.v destination *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"astounded.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"astounds.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"attempts.v theme *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"attending.v location-time *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"attending.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"awaited.v object *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"awaiting.v object *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"babbled.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"babbling.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"back.v end-time *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"back.v theme *.n" )==1)) addnote2(sock,"placing","agent",rnode);
 if((len==1)&& (matchlike(vec,"backed.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"backed.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"baffled.v attribute *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"baffled.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"baffled.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"banged.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"banged.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"baring.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"barked.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"batted.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"bawled.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 /*
 if((len==1)&& (matchlike(vec,"be.v location *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"be.v object *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"be.v object *.n" )==1)) addnote2(sock,"duplication","original",rnode);
 if((len==1)&& (matchlike(vec,"be.v object *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"be.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"be.v object *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"be.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"be.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"be.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"be.v object *.n" )==1)) addnote2(sock,"transportation","driver",rnode);
 if((len==1)&& (matchlike(vec,"be.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"be.v theme *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"be.v theme *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"be.v theme *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 */
   if((len==1)&& (matchlike(vec,"beamed.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"beat.v object *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"beat.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"beaten.v attribute *.n" )==1)) addnote2(sock,"transportation","path",rnode);
 if((len==1)&& (matchlike(vec,"beckoned.v location *.n" )==1)) addnote2(sock,"gesture","message",rnode);
 if((len==1)&& (matchlike(vec,"becoming.v object *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"beg.v theme *.n" )==1)) addnote2(sock,"request","message",rnode);
 /*
 if((len==1)&& (matchlike(vec,"began.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"began.v theme *.n" )==1)) addnote2(sock,"cogitation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"began.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"being.v object *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"being.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"being.v object *.n" )==1)) addnote2(sock,"experiencer-subj","reason",rnode);
 if((len==1)&& (matchlike(vec,"being.v object *.n" )==1)) addnote2(sock,"hear","topic",rnode);
 if((len==1)&& (matchlike(vec,"being.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"being.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 */
 if((len==1)&& (matchlike(vec,"belittle.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"bellowed.v coagent-insturment *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"bend.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"bend.v theme *.n" )==1)) addnote2(sock,"posture","agent",rnode);
 if((len==1)&& (matchlike(vec,"bent.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"bewildered.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"bewitch.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"bewitching.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"bisected.v object *.n" )==1)) addnote2(sock,"separation","whole",rnode);
 if((len==1)&& (matchlike(vec,"bit.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"bit.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"blame.v object *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"blame.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"blames.v location-time *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"blushed.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"blustered[!].v theme *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"bobbed.v location *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"boiled.v coagent-insturment *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"boiling.v coagent-insturment *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"boiling.v object *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"boiling.v object *.n" )==1)) addnote2(sock,"heat","cause",rnode);
 if((len==1)&& (matchlike(vec,"boils.v coagent-insturment *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"boils.v theme *.n" )==1)) addnote2(sock,"heat","location",rnode);
 if((len==1)&& (matchlike(vec,"bolted.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"boomed.v theme *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"booming.v attribute *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"bought.v method *.n" )==1)) addnote2(sock,"commerce","buyer",rnode);
 if((len==1)&& (matchlike(vec,"bought.v object *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"bowed.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"breaking.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"breeds.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"breeze.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"bring.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"bring.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"bring.v recipient *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"broken.v attribute *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"broken.v attribute *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"broken.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"brooded.v location-time *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"brooded.v medium *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"brooding.v destination *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"brought.v destination *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"burbled.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"burn.v object *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"burning.v object *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"bustled.v destination *.n" )==1)) addnote2(sock,"removing","goal",rnode);
 if((len==1)&& (matchlike(vec,"bustled.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"buy.v object *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"cackled.v location-time *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"called.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"calling.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"calm.v object *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"calmed.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"came.v recipient *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"came.v source *.n" )==1)) addnote2(sock,"arriving","source",rnode);
 if((len==1)&& (matchlike(vec,"came.v theme *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"came.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"came.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 /*
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"awareness","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"coming-to-believe","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"differentiation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"duplication","creator",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"duplication","original",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"emotion_active","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"hear","speaker",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"replacement","agent",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"can.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 */
 if((len==1)&& (matchlike(vec,"canter.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"capered.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"captivate.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"captivate.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"captivated.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"carpeted.v attribute *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"carried.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"caught.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"caused.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"caused.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"caused.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"cautioned.v contact *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"cautioned.v object *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"cautioned.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"ceases.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"censured.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"chafe.v theme *.n" )==1)) addnote2(sock,"heat","cause",rnode);
 if((len==1)&& (matchlike(vec,"chained.v destination *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"chair.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"chanted.v object *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"characterized.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"characterized.v theme *.n" )==1)) addnote2(sock,"categorization","criterion",rnode);
 if((len==1)&& (matchlike(vec,"characterized.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"charge.v object *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"charged.v object *.n" )==1)) addnote2(sock,"emptying","theme",rnode);
 if((len==1)&& (matchlike(vec,"charged.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"charmed.v coagent-insturment *.n" )==1)) addnote2(sock,"experiencer-obj","means",rnode);
 if((len==1)&& (matchlike(vec,"charmed.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"chase.v object *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"chase.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"chased.v location *.n" )==1)) addnote2(sock,"cotheme","manner",rnode);
 if((len==1)&& (matchlike(vec,"chased.v method *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"chased.v object *.n" )==1)) addnote2(sock,"cotheme","cotheme",rnode);
 if((len==1)&& (matchlike(vec,"chased.v object *.n" )==1)) addnote2(sock,"cotheme","path",rnode);
 if((len==1)&& (matchlike(vec,"chastising.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"chat.v theme *.n" )==1)) addnote2(sock,"conversation","interlocutors",rnode);
 if((len==1)&& (matchlike(vec,"chatter.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"chattered.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"chattered.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"chattering.v attribute *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"chattering.v attribute *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"chattering.v attribute *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"chattering.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"cheered.v attribute *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"cheered.v destination *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"cheered.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"cheered.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"chirped.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"chirruped.v recipient *.n" )==1)) addnote2(sock,"communication_noise","addressee",rnode);
 if((len==1)&& (matchlike(vec,"chuck.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"chucked.v destination *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"chuckling.v coagent-insturment *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"chucks.v location *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"chucks.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"chugs.v theme *.n" )==1)) addnote2(sock,"motion_noise","source",rnode);
 if((len==1)&& (matchlike(vec,"circle.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"claim.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"claim.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"claimed.v method *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"claimed.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"claimed.v theme *.n" )==1)) addnote2(sock,"transportation","driver",rnode);
 if((len==1)&& (matchlike(vec,"claims.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"clamber.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"clambered.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"clanged.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"clanged.v theme *.n" )==1)) addnote2(sock,"motion_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"clanged.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"clanged.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"clapped.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"clattered.v location-time *.n" )==1)) addnote2(sock,"motion_noise","goal",rnode);
 if((len==1)&& (matchlike(vec,"clattered.v medium *.n" )==1)) addnote2(sock,"motion_noise","goal",rnode);
 if((len==1)&& (matchlike(vec,"clattered.v source *.n" )==1)) addnote2(sock,"motion_noise","source",rnode);
 if((len==1)&& (matchlike(vec,"clattered.v source *.n" )==1)) addnote2(sock,"perception_noise","source",rnode);
 if((len==1)&& (matchlike(vec,"clattering.v location *.n" )==1)) addnote2(sock,"motion_noise","goal",rnode);
 if((len==1)&& (matchlike(vec,"clean.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"clear.v object *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"clear.v object *.n" )==1)) addnote2(sock,"emptying","theme",rnode);
 if((len==1)&& (matchlike(vec,"clear.v object *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"clear.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"cleared.v theme *.n" )==1)) addnote2(sock,"emptying","agent",rnode);
 if((len==1)&& (matchlike(vec,"cleared.v theme *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"clench.v object *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"clench.v object *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"clench.v object *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"clicking.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"climbed.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"climbed.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"climbing.v object *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"clinked.v location *.n" )==1)) addnote2(sock,"motion_noise","goal",rnode);
 if((len==1)&& (matchlike(vec,"clinked.v recipient *.n" )==1)) addnote2(sock,"motion_noise","goal",rnode);
 if((len==1)&& (matchlike(vec,"clumped.v object *.n" )==1)) addnote2(sock,"motion_noise","path",rnode);
 if((len==1)&& (matchlike(vec,"clumped.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"clunked.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"clutched.v theme *.n" )==1)) addnote2(sock,"manipulation","agent",rnode);
 if((len==1)&& (matchlike(vec,"clutching.v object *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"cobbled.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"cocked.v location-time *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"cocked.v object *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"cocked.v recipient *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"collapsed.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"combine.v object *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"combined.v theme *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"come.v manner *.n" )==1)) addnote2(sock,"arriving","path",rnode);
 if((len==1)&& (matchlike(vec,"come.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"comfort.v location *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"comfort.v location-time *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"comfort.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"comfort.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"comforted.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"coming.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"commanded.v object *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"commanded.v theme *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"comments.v coagent-insturment *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 if((len==1)&& (matchlike(vec,"comments.v recipient *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"comprehend.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"comprehend.v theme *.n" )==1)) addnote2(sock,"awareness","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"conceive.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"conceive.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"conceived.v method *.n" )==1)) addnote2(sock,"awareness","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"conceived.v method *.n" )==1)) addnote2(sock,"invention","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"conceived.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"conceived.v theme *.n" )==1)) addnote2(sock,"awareness","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"conceived.v theme *.n" )==1)) addnote2(sock,"coming-to-believe","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"conceived.v theme *.n" )==1)) addnote2(sock,"invention","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"concoct.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"concocted.v method *.n" )==1)) addnote2(sock,"invention","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"concocted.v theme *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"concocting.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"concocting.v theme *.n" )==1)) addnote2(sock,"invention","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"condemn.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"condemned.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"condemned.v object *.n" )==1)) addnote2(sock,"judgment","role",rnode);
 if((len==1)&& (matchlike(vec,"condemned.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"condemning.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"condemns.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"confer.v coagent-insturment *.n" )==1)) addnote2(sock,"conversation","interlocutor-2",rnode);
 if((len==1)&& (matchlike(vec,"confer.v theme *.n" )==1)) addnote2(sock,"conversation","interlocutor-1",rnode);
 if((len==1)&& (matchlike(vec,"conferred.v theme *.n" )==1)) addnote2(sock,"conversation","interlocutors",rnode);
 if((len==1)&& (matchlike(vec,"confessed.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"confirm.v theme *.n" )==1)) addnote2(sock,"evidence","support",rnode);
 if((len==1)&& (matchlike(vec,"confiscated.v theme *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"confiscated.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"confused.v agent *.n" )==1)) addnote2(sock,"body-movement","cause",rnode);
 if((len==1)&& (matchlike(vec,"confused.v agent *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"confused.v attribute *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"consent.v destination *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"consent.v theme *.n" )==1)) addnote2(sock,"commitment","speaker",rnode);
 if((len==1)&& (matchlike(vec,"considering.v object *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"considering.v theme *.n" )==1)) addnote2(sock,"cogitation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"contemplated.v object *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"contemplating.v object *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"contemplating.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"continued.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"contrive.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"control.v object *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"converse.v location-time *.n" )==1)) addnote2(sock,"conversation","medium",rnode);
 if((len==1)&& (matchlike(vec,"copy.v location *.n" )==1)) addnote2(sock,"duplication","goal",rnode);
 if((len==1)&& (matchlike(vec,"copy.v object *.n" )==1)) addnote2(sock,"duplication","original",rnode);
 if((len==1)&& (matchlike(vec,"cost.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"communication_response","speaker",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"cure","patient",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"imitation","creator",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"replacement","agents",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"could.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"cover.v coagent-insturment *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"cover.v coagent-insturment *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"covered.v agent *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"covered.v agent *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"covered.v coagent *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"covered.v coagent *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"covered.v method *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"covered.v object *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"covered.v object *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"covered.v theme *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"covered.v theme *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"cowled[!].v attribute *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"cramped.v attribute *.n" )==1)) addnote2(sock,"communication_noise","topic",rnode);
 if((len==1)&& (matchlike(vec,"craned.v object *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"craned.v object *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"craned.v object *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"crashed.v agent *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"craved.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"craved.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"craves.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"crawled.v recipient *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"crawling.v attribute *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"create.v object *.n" )==1)) addnote2(sock,"expansion","agent",rnode);
 if((len==1)&& (matchlike(vec,"created.v source *.n" )==1)) addnote2(sock,"invention","material",rnode);
 if((len==1)&& (matchlike(vec,"created.v theme *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"creates.v coagent-insturment *.n" )==1)) addnote2(sock,"invention","material",rnode);
 if((len==1)&& (matchlike(vec,"creates.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"crept.v destination *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"crept.v recipient *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"crept.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"crisscross.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"croaked.v location-time *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"croaked.v medium *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"crossed.v location *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"crossed.v location *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"crossed.v object *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"crossed.v object *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"crossed.v recipient *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"crossed.v recipient *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"crowed.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"crowned.v method *.n" )==1)) addnote2(sock,"filling","agent",rnode);
 if((len==1)&& (matchlike(vec,"cruised.v theme *.n" )==1)) addnote2(sock,"transportation","driver",rnode);
 if((len==1)&& (matchlike(vec,"cruising.v location *.n" )==1)) addnote2(sock,"transportation","area",rnode);
 if((len==1)&& (matchlike(vec,"crumpled.v location *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"crunched.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"cured.v coagent-insturment *.n" )==1)) addnote2(sock,"cure","patient",rnode);
 if((len==1)&& (matchlike(vec,"cured.v object *.n" )==1)) addnote2(sock,"cure","patient",rnode);
 if((len==1)&& (matchlike(vec,"curved.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"curved.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"cycle.v agent *.n" )==1)) addnote2(sock,"transportation","path",rnode);
 if((len==1)&& (matchlike(vec,"danced.v location-time *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"dances.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"dart.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"darted.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"dashed.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"dashed.v object *.n" )==1)) addnote2(sock,"arriving","path",rnode);
 if((len==1)&& (matchlike(vec,"dashed.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"dazzled.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"decamped.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"declared.v destination *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"declared.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"declared.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"decorate.v object *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"decorate.v theme *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"decorating.v object *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"deducing.v object *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"defend.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"define.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"defined.v path *.n" )==1)) addnote2(sock,"categorization","criterion",rnode);
 if((len==1)&& (matchlike(vec,"defines.v location-time *.n" )==1)) addnote2(sock,"categorization","criterion",rnode);
 if((len==1)&& (matchlike(vec,"defining.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"delighted.v attribute *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"delighted.v attribute *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"delighted.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"demonstrate.v object *.n" )==1)) addnote2(sock,"evidence","proposition",rnode);
 if((len==1)&& (matchlike(vec,"demonstrates.v theme *.n" )==1)) addnote2(sock,"evidence","support",rnode);
 if((len==1)&& (matchlike(vec,"denigrated.v location *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"denigrated.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"denigrating.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"denounced.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"denounces.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"depart.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"departed.v coagent-insturment *.n" )==1)) addnote2(sock,"departing","manner",rnode);
 if((len==1)&& (matchlike(vec,"departed.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"departed.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"departing.v source *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"departing.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"depends.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"depends.v theme *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"depicted.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"deplore.v location-time *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"deplore.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"deplored.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"deplored.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"depose.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"deposed.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"deposited.v method *.n" )==1)) addnote2(sock,"placing","agent",rnode);
 if((len==1)&& (matchlike(vec,"deposited.v method *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"deposited.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"deposited.v theme *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"deposited.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"deprecate.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"deprecating.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"depress.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"derided.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"deriding.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"descend.v location-time *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"descended.v destination *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"descended.v object *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"descends.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"described.v theme *.n" )==1)) addnote2(sock,"categorization","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"desert.v location *.n" )==1)) addnote2(sock,"departing","manner",rnode);
 if((len==1)&& (matchlike(vec,"desert.v location-time *.n" )==1)) addnote2(sock,"departing","manner",rnode);
 if((len==1)&& (matchlike(vec,"desert.v object *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"deserted.v object *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"deserted.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"designs.v coagent-insturment *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"designs.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"desired.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"desires.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"desires.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"despaired.v location *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"despise.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"despise.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"despised.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"despises.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"detect.v location-time *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"detect.v medium *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"detect.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"detected.v location-time *.n" )==1)) addnote2(sock,"perception-experience","ground",rnode);
 if((len==1)&& (matchlike(vec,"detected.v medium *.n" )==1)) addnote2(sock,"perception-experience","ground",rnode);
 if((len==1)&& (matchlike(vec,"detested.v coagent-insturment *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"detested.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"devised.v destination *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"devised.v theme *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"did.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"did.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"did.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"died.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"differentiate.v object *.n" )==1)) addnote2(sock,"differentiation","phenomena",rnode);
 if((len==1)&& (matchlike(vec,"digested.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"dilate.v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"dilate.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"diminish.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"dip.v location *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"dip.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"dipped.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"dipped.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"disabling.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"disappeared.v location-time *.n" )==1)) addnote2(sock,"departing","path",rnode);
 if((len==1)&& (matchlike(vec,"disappointed.v theme *.n" )==1)) addnote2(sock,"expectation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"discarded.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"discarded.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"discarded.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"disclosed.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"disclosing.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"discomfited.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"discomfited.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"discomfited.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"discover.v coagent-insturment *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"discover.v object *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"discuss.v object *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"discussed.v coagent *.n" )==1)) addnote2(sock,"conversation","interlocutor-2",rnode);
 if((len==1)&& (matchlike(vec,"discussed.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"discussing.v coagent-insturment *.n" )==1)) addnote2(sock,"conversation","interlocutor-2",rnode);
 if((len==1)&& (matchlike(vec,"discussing.v object *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"discussing.v object *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 if((len==1)&& (matchlike(vec,"disdain.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"disdained.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"dislike.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"disliked.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"disliked.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"disliking.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"dislodged.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"dislodged.v theme *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"dislodged.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"dismissed.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"dismissing.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"disparage.v location-time *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"displeased.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"dissatisfied[!].v coagent *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"distinguishes.v object *.n" )==1)) addnote2(sock,"differentiation","phenomena",rnode);
 if((len==1)&& (matchlike(vec,"distressed.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"do.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"do.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"do.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"do.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"do.v theme *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"do.v theme *.n" )==1)) addnote2(sock,"recovery","body part",rnode);
 if((len==1)&& (matchlike(vec,"do.v theme *.n" )==1)) addnote2(sock,"transportation","driver",rnode);
 if((len==1)&& (matchlike(vec,"does.v theme *.n" )==1)) addnote2(sock,"evidence","support",rnode);
 if((len==1)&& (matchlike(vec,"does.v theme *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 if((len==1)&& (matchlike(vec,"does.v theme *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"does.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"doing.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"doing.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"doing.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"doing.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"dragged.v location *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"dragged.v object *.n" )==1)) addnote2(sock,"cause-to-move","path",rnode);
 if((len==1)&& (matchlike(vec,"dragged.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"dragged.v theme *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"drain.v object *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"drain.v object *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"drained.v object *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"drained.v object *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"drains.v object *.n" )==1)) addnote2(sock,"emptying","theme",rnode);
 if((len==1)&& (matchlike(vec,"drains.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"drains.v source *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"drains.v source *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"dreaded.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"dreading.v coagent *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"drink.v theme *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"driven.v theme *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"driving.v object *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 if((len==1)&& (matchlike(vec,"droned.v theme *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v destination *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v destination *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v location *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v location *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v location-time *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v location-time *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v medium *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v medium *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v object *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v recipient *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v recipient *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"dropped.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"drops.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"drove.v object *.n" )==1)) addnote2(sock,"transportation","distance",rnode);
 if((len==1)&& (matchlike(vec,"drowned.v attribute *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"duplicated.v location *.n" )==1)) addnote2(sock,"duplication","goal",rnode);
 if((len==1)&& (matchlike(vec,"dwell.v theme *.n" )==1)) addnote2(sock,"cogitation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"eating.v object *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"eating.v object *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 if((len==1)&& (matchlike(vec,"edged.v agent *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"edged.v destination *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"edged.v location *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"edged.v method *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"edged.v object *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"edged.v theme *.n" )==1)) addnote2(sock,"emptying","path",rnode);
 if((len==1)&& (matchlike(vec,"edged.v theme *.n" )==1)) addnote2(sock,"removing","path",rnode);
 if((len==1)&& (matchlike(vec,"edged.v theme *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"eject.v source *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"ejected.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"eliminate.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"embarrass.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"emerged.v source *.n" )==1)) addnote2(sock,"path-shape","source",rnode);
 if((len==1)&& (matchlike(vec,"emigrated.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"emptied.v object *.n" )==1)) addnote2(sock,"placing","source",rnode);
 if((len==1)&& (matchlike(vec,"enchanted.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"encircled.v agent *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"encircled.v coagent-insturment *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"encircled.v method *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"enclose.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"end.v object *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"end.v theme *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"enjoy.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"enjoy.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"enlarge.v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"enlarge.v theme *.n" )==1)) addnote2(sock,"expansion","agent",rnode);
 if((len==1)&& (matchlike(vec,"enlarged.v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"enlarged.v theme *.n" )==1)) addnote2(sock,"expansion","agent",rnode);
 if((len==1)&& (matchlike(vec,"enraged.v attribute *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"enraged.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"enter.v object *.n" )==1)) addnote2(sock,"arriving","goal",rnode);
 if((len==1)&& (matchlike(vec,"enter.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"entered.v source *.n" )==1)) addnote2(sock,"arriving","source",rnode);
 if((len==1)&& (matchlike(vec,"entered.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"entertaining.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"enthralled.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"enveloped.v method *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"enveloped.v theme *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"enveloped.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"enveloped.v theme *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"envy.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"envy.v object *.n" )==1)) addnote2(sock,"experiencer-subj","reason",rnode);
 if((len==1)&& (matchlike(vec,"escape.v start-time *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"escaped.v location-time *.n" )==1)) addnote2(sock,"departing","manner",rnode);
 if((len==1)&& (matchlike(vec,"escaped.v object *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"established.v attribute *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"evacuate.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"evict.v source *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"examine.v theme *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"exasperated.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"exchange.v object *.n" )==1)) addnote2(sock,"replacement","places",rnode);
 if((len==1)&& (matchlike(vec,"exchanged.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"exchanged.v theme *.n" )==1)) addnote2(sock,"replacement","themes",rnode);
 if((len==1)&& (matchlike(vec,"excised.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"excite.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"excited.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"excused.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"executed.v theme *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"exemplify.v theme *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"existing.v attribute *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"expand.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"expanding.v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"expanding.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"expected.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"expected.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"expected.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"expecting.v theme *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"expects.v location *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"expects.v location-time *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"explained.v object *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"explained.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"explore.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"express.v object *.n" )==1)) addnote2(sock,"encoding","message",rnode);
 if((len==1)&& (matchlike(vec,"expresses.v theme *.n" )==1)) addnote2(sock,"encoding","medium",rnode);
 if((len==1)&& (matchlike(vec,"expunged.v source *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"expunged.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"exterminated.v theme *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"extolled.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"extols.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"fake.v theme *.n" )==1)) addnote2(sock,"imitation","creator",rnode);
 if((len==1)&& (matchlike(vec,"fallen.v attribute *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"falsify.v object *.n" )==1)) addnote2(sock,"imitation","original",rnode);
 if((len==1)&& (matchlike(vec,"fascinate.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"fascinate.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"fascinated.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"fascinated.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"fault.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"fear.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"feared.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"feared.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"feared.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"fears.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"feel.v object *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"feel.v object *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"feel.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"feel.v theme *.n" )==1)) addnote2(sock,"encoding","message",rnode);
 if((len==1)&& (matchlike(vec,"feign.v object *.n" )==1)) addnote2(sock,"imitation","original",rnode);
 if((len==1)&& (matchlike(vec,"feign.v theme *.n" )==1)) addnote2(sock,"imitation","creator",rnode);
 if((len==1)&& (matchlike(vec,"fell.v source *.n" )==1)) addnote2(sock,"self-motion","manner",rnode);
 if((len==1)&& (matchlike(vec,"fell.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"fell.v theme *.n" )==1)) addnote2(sock,"self-motion","manner",rnode);
 if((len==1)&& (matchlike(vec,"felt.v location *.n" )==1)) addnote2(sock,"departing","goal",rnode);
 if((len==1)&& (matchlike(vec,"felt.v location-time *.n" )==1)) addnote2(sock,"appearance","state",rnode);
 if((len==1)&& (matchlike(vec,"felt.v location-time *.n" )==1)) addnote2(sock,"perception-experience","state",rnode);
 if((len==1)&& (matchlike(vec,"felt.v location-time *.n" )==1)) addnote2(sock,"perception_active","state",rnode);
 if((len==1)&& (matchlike(vec,"felt.v medium *.n" )==1)) addnote2(sock,"appearance","state",rnode);
 if((len==1)&& (matchlike(vec,"felt.v medium *.n" )==1)) addnote2(sock,"perception-experience","state",rnode);
 if((len==1)&& (matchlike(vec,"felt.v medium *.n" )==1)) addnote2(sock,"perception_active","state",rnode);
 if((len==1)&& (matchlike(vec,"fighting.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"fill.v object *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"fill.v theme *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"filmed.v theme *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"find.v location *.n" )==1)) addnote2(sock,"becoming-aware","ground",rnode);
 if((len==1)&& (matchlike(vec,"find.v object *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"find.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"fingered.v coagent *.n" )==1)) addnote2(sock,"manipulation","agent's body part",rnode);
 if((len==1)&& (matchlike(vec,"fingered.v coagent-insturment *.n" )==1)) addnote2(sock,"manipulation","agent's body part",rnode);
 if((len==1)&& (matchlike(vec,"finish.v coagent *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"fire.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"fizz.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"flapped.v location *.n" )==1)) addnote2(sock,"body-movement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"flapped.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"flapping.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"flaps.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"flaps.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"flared.v attribute *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"flew.v object *.n" )==1)) addnote2(sock,"transportation","path",rnode);
 if((len==1)&& (matchlike(vec,"flew.v theme *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 if((len==1)&& (matchlike(vec,"fling.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"fling.v source *.n" )==1)) addnote2(sock,"body-movement","source",rnode);
 if((len==1)&& (matchlike(vec,"fling.v source *.n" )==1)) addnote2(sock,"cause-to-move","source",rnode);
 if((len==1)&& (matchlike(vec,"flinging.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"flown.v object *.n" )==1)) addnote2(sock,"transportation","distance",rnode);
 if((len==1)&& (matchlike(vec,"flown.v theme *.n" )==1)) addnote2(sock,"transportation","cargo + passenger",rnode);
 if((len==1)&& (matchlike(vec,"flown.v theme *.n" )==1)) addnote2(sock,"transportation","driver",rnode);
 if((len==1)&& (matchlike(vec,"flung.v attribute *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"flung.v recipient *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"flung.v recipient *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"flushed.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"flushed.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"fluttered.v location *.n" )==1)) addnote2(sock,"body-movement","area",rnode);
 if((len==1)&& (matchlike(vec,"fluttered.v location-time *.n" )==1)) addnote2(sock,"body-movement","area",rnode);
 if((len==1)&& (matchlike(vec,"fly.v object *.n" )==1)) addnote2(sock,"transportation","cargo + passenger",rnode);
 if((len==1)&& (matchlike(vec,"fly.v object *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 if((len==1)&& (matchlike(vec,"fly.v origin *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"fly.v origin *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"fly.v origin *.n" )==1)) addnote2(sock,"transportation","source",rnode);
 if((len==1)&& (matchlike(vec,"followed.v agent *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"followed.v location *.n" )==1)) addnote2(sock,"cotheme","goal",rnode);
 if((len==1)&& (matchlike(vec,"followed.v theme *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"force.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"forces.v object *.n" )==1)) addnote2(sock,"replacement","agent",rnode);
 if((len==1)&& (matchlike(vec,"foreseen.v theme *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"form.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"forming.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"formulate.v object *.n" )==1)) addnote2(sock,"encoding","message",rnode);
 if((len==1)&& (matchlike(vec,"found.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"found.v object *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"found.v object *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"freckled[!].v attribute *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"fret.v theme *.n" )==1)) addnote2(sock,"emotion_active","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"fretted.v theme *.n" )==1)) addnote2(sock,"emotion_active","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"fretted.v theme *.n" )==1)) addnote2(sock,"heat","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"fried.v attribute *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"frowned.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"frowning.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"frozen.v attribute *.n" )==1)) addnote2(sock,"motion_noise","path",rnode);
 if((len==1)&& (matchlike(vec,"frustrate.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"frustrated.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"fume.v theme *.n" )==1)) addnote2(sock,"heat","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"gambolling.v location-time *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"gasped.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"gave.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"gave.v object *.n" )==1)) addnote2(sock,"judgment","expressor",rnode);
 if((len==1)&& (matchlike(vec,"gaze.v theme *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 if((len==1)&& (matchlike(vec,"gazed.v location *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"gazed.v location *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"gazed.v location-time *.n" )==1)) addnote2(sock,"perception_active","manner",rnode);
 if((len==1)&& (matchlike(vec,"gazed.v source *.n" )==1)) addnote2(sock,"perception_active","location of perceiver",rnode);
 if((len==1)&& (matchlike(vec,"gazing.v location *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"gestured.v theme *.n" )==1)) addnote2(sock,"gesture","communicator",rnode);
 if((len==1)&& (matchlike(vec,"get.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"get.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"get.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"get.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"gets.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"gets.v theme *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"getting.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"give.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"given.v destination *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"gladdened.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"gladdened.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"glanced.v coagent-insturment *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"glanced.v location-time *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"glanced.v medium *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"glanced.v recipient *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"gleaming.v attribute *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"glint.v coagent *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"glint.v theme *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"glittering.v attribute *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"glowed.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"go.v recipient *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"go.v recipient *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"go.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"go.v theme *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"got.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"got.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"got.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"got.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"govern.v object *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"grasped.v object *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"grimaced.v location *.n" )==1)) addnote2(sock,"body-movement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"grimaced.v location *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"grimaced.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"grimacing.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"grinned.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"grinned.v location-time *.n" )==1)) addnote2(sock,"path-shape","manner",rnode);
 if((len==1)&& (matchlike(vec,"grinned.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"grinning.v coagent-insturment *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"grins.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"gripped.v object *.n" )==1)) addnote2(sock,"manipulation","locus",rnode);
 if((len==1)&& (matchlike(vec,"groan.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"groan.v object *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"groaned.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"ground.v attribute *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"grow.v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"grow.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"growled.v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"grows.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"grunted.v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"grunted.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"grunts.v coagent-insturment *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"grunts.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"guess.v object *.n" )==1)) addnote2(sock,"coming-to-believe","topic",rnode);
 if((len==1)&& (matchlike(vec,"guessed.v object *.n" )==1)) addnote2(sock,"coming-to-believe","topic",rnode);
 if((len==1)&& (matchlike(vec,"guided.v start-time *.n" )==1)) addnote2(sock,"cotheme","source",rnode);
 if((len==1)&& (matchlike(vec,"gusted[!].v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"gusted[!].v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
/*
 if((len==1)&& (matchlike(vec,"had.v object *.n" )==1)) addnote2(sock,"becoming-aware","ground",rnode);
 if((len==1)&& (matchlike(vec,"had.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"had.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"commitment","speaker",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"emptying","agent",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"motion_noise","path",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"separation","whole",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"had.v theme *.n" )==1)) addnote2(sock,"transportation","cargo + passenger",rnode);
 */
 if((len==1)&& (matchlike(vec,"halted.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"hang.v location-time *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"hang.v location-time *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"happened.v destination *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 /*
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"categorization","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"categorization","criterion",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"evidence","support",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"has.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 */
 if((len==1)&& (matchlike(vec,"hastened.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"hastened.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"hastens.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"hatched.v attribute *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"hatched.v theme *.n" )==1)) addnote2(sock,"invention","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"hatched.v theme *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"hatching.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"hate.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"hated.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"hated.v medium *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"hated.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"hated.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"hates.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"haul.v location-time *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"haul.v location-time *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"hauled.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"hauled.v origin *.n" )==1)) addnote2(sock,"cause-to-move","source",rnode);
 if((len==1)&& (matchlike(vec,"hauled.v theme *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 /*
 if((len==1)&& (matchlike(vec,"have.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"have.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"have.v object *.n" )==1)) addnote2(sock,"cogitation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"have.v object *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"have.v object *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"have.v object *.n" )==1)) addnote2(sock,"emotion_active","topic",rnode);
 if((len==1)&& (matchlike(vec,"have.v object *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"have.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"coming-to-believe","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"encoding","speaker",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"replacement","theme-old",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"request","addressee",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"have.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 */
 if((len==1)&& (matchlike(vec,"having.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"head.v object *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"head.v theme *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"heal.v location *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"heal.v location *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"heal.v location-time *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"heal.v location-time *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"heal.v object *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"heal.v object *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"heal.v theme *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"heal.v theme *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"heal.v theme *.n" )==1)) addnote2(sock,"recovery","body part",rnode);
 if((len==1)&& (matchlike(vec,"hear.v location-time *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"hear.v location-time *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"hear.v location-time *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"hear.v medium *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"hear.v object *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"heard.v agent *.n" )==1)) addnote2(sock,"perception-experience","perceiver-passive",rnode);
 if((len==1)&& (matchlike(vec,"heard.v location *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"heard.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"heard.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"heard.v source *.n" )==1)) addnote2(sock,"hear","speaker",rnode);
 if((len==1)&& (matchlike(vec,"heard.v source *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"heard.v source *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"heard.v theme *.n" )==1)) addnote2(sock,"duplication","goal",rnode);
 if((len==1)&& (matchlike(vec,"heard.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"heartened.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"heartened.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"heartened.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"hissed.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"hit.v theme *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"hobbled.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"hold.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"hold.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"hounded.v method *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"howled.v location-time *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"howled.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"humiliate.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"humiliated.v agent *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"hung.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"hurl.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"hurl.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"hurled.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"hurt.v coagent-insturment *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"hurt.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"hurt.v object *.n" )==1)) addnote2(sock,"body","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"hurting.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"hurts.v object *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"hyperventilate[?].v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"identify.v location-time *.n" )==1)) addnote2(sock,"categorization","criterion",rnode);
 if((len==1)&& (matchlike(vec,"implicated.v theme *.n" )==1)) addnote2(sock,"evidence","proposition",rnode);
 if((len==1)&& (matchlike(vec,"implored.v object *.n" )==1)) addnote2(sock,"request","addressee",rnode);
 if((len==1)&& (matchlike(vec,"implored.v object *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"implored.v theme *.n" )==1)) addnote2(sock,"request","addressee",rnode);
 if((len==1)&& (matchlike(vec,"implored.v theme *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"impressed.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"impressed.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"improvise.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"included.v theme *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"includes.v object *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"increased.v attribute *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"increased.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"increasing.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"indicated.v theme *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"infected.v attribute *.n" )==1)) addnote2(sock,"evidence","proposition",rnode);
 if((len==1)&& (matchlike(vec,"inflated.v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"inflated.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"informed.v object *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"infuriate.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"infuriated.v agent *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"initiating.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"inquire.v manner *.n" )==1)) addnote2(sock,"questioning","message",rnode);
 if((len==1)&& (matchlike(vec,"inserted.v location *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"inserted.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"installed.v location *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"insulting.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"intended.v attribute *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"interchanged.v theme *.n" )==1)) addnote2(sock,"replacement","themes",rnode);
 if((len==1)&& (matchlike(vec,"interpreted.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"interrogated.v agent *.n" )==1)) addnote2(sock,"questioning","speaker",rnode);
 if((len==1)&& (matchlike(vec,"interviewed.v object *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"invading.v object *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"invent.v object *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"invited.v object *.n" )==1)) addnote2(sock,"replacement","agents",rnode);
 if((len==1)&& (matchlike(vec,"involved.v location-time *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"irked.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"irked.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"irritated.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"irritated.v theme *.n" )==1)) addnote2(sock,"causation","cause",rnode);
 if((len==1)&& (matchlike(vec,"irritated.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 /*
 if((len==1)&& (matchlike(vec,"is.v location-time *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"is.v location-time *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"is.v medium *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"is.v object *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"is.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"is.v object *.n" )==1)) addnote2(sock,"imitation","original",rnode);
 if((len==1)&& (matchlike(vec,"is.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"is.v object *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"is.v object *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"arriving","goal",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"change-of-leadership","new leader",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"cogitation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"emptying","path",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"encoding","message",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"expansion","agent",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"invention","invention",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"removing","path",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"is.v theme *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 */
 if((len==1)&& (matchlike(vec,"jabbered.v theme *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"jerked.v coagent-insturment *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"jerked.v location *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"jesting.v manner *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"jiggled.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"joked.v coagent *.n" )==1)) addnote2(sock,"conversation","interlocutor-2",rnode);
 if((len==1)&& (matchlike(vec,"joked.v location *.n" )==1)) addnote2(sock,"conversation","interlocutor-2",rnode);
 if((len==1)&& (matchlike(vec,"joked.v location-time *.n" )==1)) addnote2(sock,"conversation","interlocutor-2",rnode);
 if((len==1)&& (matchlike(vec,"joking.v coagent-insturment *.n" )==1)) addnote2(sock,"conversation","interlocutor-2",rnode);
 if((len==1)&& (matchlike(vec,"jumped.v destination *.n" )==1)) addnote2(sock,"manner","addressee",rnode);
 if((len==1)&& (matchlike(vec,"jumped.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"jumped.v location *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"jumped.v location-time *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"jumped.v object *.n" )==1)) addnote2(sock,"self-motion","distance",rnode);
 if((len==1)&& (matchlike(vec,"jumped.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"jumped.v object *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"jumped.v recipient *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"jumped.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"jumping.v source *.n" )==1)) addnote2(sock,"body-movement","source",rnode);
 if((len==1)&& (matchlike(vec,"jumping.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"jumping.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"jumping.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"kept.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"killing.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"kneaded.v object *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"kneeling.v location *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"kneeling.v location-time *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"kneels.v agent *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"knelt.v agent *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"knelt.v location-time *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"knelt.v medium *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"knelt.v theme *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"knelt.v theme *.n" )==1)) addnote2(sock,"posture","agent",rnode);
 if((len==1)&& (matchlike(vec,"knew.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"know.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"know.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"laid.v theme *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"lead.v object *.n" )==1)) addnote2(sock,"cotheme","cotheme",rnode);
 if((len==1)&& (matchlike(vec,"leading.v object *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"leaned.v contact *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"leaned.v destination *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"leaned.v location-time *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"leaned.v theme *.n" )==1)) addnote2(sock,"placing","agent",rnode);
 if((len==1)&& (matchlike(vec,"leaned.v theme *.n" )==1)) addnote2(sock,"posture","agent",rnode);
 if((len==1)&& (matchlike(vec,"leap.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"leaping.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"leapt.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"leapt.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"learn.v object *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"learn.v theme *.n" )==1)) addnote2(sock,"coming-to-believe","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"learned.v location *.n" )==1)) addnote2(sock,"coming-to-believe","evidence",rnode);
 if((len==1)&& (matchlike(vec,"learned.v method *.n" )==1)) addnote2(sock,"coming-to-believe","evidence",rnode);
 if((len==1)&& (matchlike(vec,"learned.v source *.n" )==1)) addnote2(sock,"coming-to-believe","evidence",rnode);
 if((len==1)&& (matchlike(vec,"learning.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"learning.v object *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"learns.v source *.n" )==1)) addnote2(sock,"coming-to-believe","evidence",rnode);
 if((len==1)&& (matchlike(vec,"leave.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"leave.v object *.n" )==1)) addnote2(sock,"path-shape","source",rnode);
 if((len==1)&& (matchlike(vec,"leaves.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"leaves.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"leaving.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"led.v destination *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"left.v object *.n" )==1)) addnote2(sock,"path-shape","source",rnode);
 if((len==1)&& (matchlike(vec,"left.v object *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"left.v theme *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"left.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"lengthen.v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"lengthened.v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"lengthens.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"let.v coagent-insturment *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"level.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"lift.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"lifted.v object *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"lifted.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"lifted.v object *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"lifted.v recipient *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"lifted.v recipient *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"lifted.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"lifts.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"like.v location *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"like.v medium *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"like.v object *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"like.v object *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"like.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"like.v object *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"like.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"liked.v location *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"limped.v destination *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"limped.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"limped.v location *.n" )==1)) addnote2(sock,"cotheme","goal",rnode);
 if((len==1)&& (matchlike(vec,"lined.v theme *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"listening.v theme *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"listening.v theme *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 if((len==1)&& (matchlike(vec,"live.v theme *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"loathe.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"loathed.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"longed.v contact *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"look.v destination *.n" )==1)) addnote2(sock,"appearance","direction",rnode);
 if((len==1)&& (matchlike(vec,"look.v destination *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"look.v location-time *.n" )==1)) addnote2(sock,"appearance","direction",rnode);
 if((len==1)&& (matchlike(vec,"look.v location-time *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"look.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"look.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"look.v theme *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"look.v theme *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"looked.v analogy *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"looked.v destination *.n" )==1)) addnote2(sock,"appearance","direction",rnode);
 if((len==1)&& (matchlike(vec,"looked.v destination *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"looked.v location *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"looked.v location *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"looked.v location *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"looked.v location-time *.n" )==1)) addnote2(sock,"appearance","state",rnode);
 if((len==1)&& (matchlike(vec,"looked.v location-time *.n" )==1)) addnote2(sock,"perception_active","state",rnode);
 if((len==1)&& (matchlike(vec,"looked.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"looking.v location-time *.n" )==1)) addnote2(sock,"appearance","direction",rnode);
 if((len==1)&& (matchlike(vec,"looking.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"looking.v location-time *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"looks.v analogy *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"looks.v theme *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"loped.v location *.n" )==1)) addnote2(sock,"self-motion","manner",rnode);
 if((len==1)&& (matchlike(vec,"losing.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"losing.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"losing.v object *.n" )==1)) addnote2(sock,"heat","cause",rnode);
 if((len==1)&& (matchlike(vec,"losing.v object *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"lost.v object *.n" )==1)) addnote2(sock,"appearance","inference",rnode);
 if((len==1)&& (matchlike(vec,"lost.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"love.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"love.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"lowered.v location *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"lurched.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"lurched.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"lurches.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"lurching.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"lurk.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"made.v location-time *.n" )==1)) addnote2(sock,"causation","cause",rnode);
 if((len==1)&& (matchlike(vec,"made.v medium *.n" )==1)) addnote2(sock,"causation","cause",rnode);
 if((len==1)&& (matchlike(vec,"made.v object *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"made.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"made.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"made.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"made.v theme *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"made.v theme *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"maintain.v object *.n" )==1)) addnote2(sock,"duplication","original",rnode);
 if((len==1)&& (matchlike(vec,"make.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"make.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"makes.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"making.v object *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"making.v source *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"marched.v destination *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"marched.v location *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"marched.v location *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"marched.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"marched.v manner *.n" )==1)) addnote2(sock,"self-motion","distance",rnode);
 if((len==1)&& (matchlike(vec,"marched.v recipient *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"marched.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"marched.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"marching.v attribute *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"maried[!].v attribute *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"maried[!].v attribute *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"maried[!].v attribute *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"marked.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"marked.v recipient *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"marked.v theme *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"massaged.v location *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"massaged.v theme *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"may.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"may.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"may.v theme *.n" )==1)) addnote2(sock,"duplication","original",rnode);
 if((len==1)&& (matchlike(vec,"may.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"may.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"may.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"may.v theme *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"may.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"meant.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"meant.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"meant.v theme *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"meant.v theme *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"merge.v theme *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"merged.v theme *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"merging.v location *.n" )==1)) addnote2(sock,"joining","whole",rnode);
 if((len==1)&& (matchlike(vec,"might.v theme *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"might.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"might.v theme *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 if((len==1)&& (matchlike(vec,"minced.v object *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"mine.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"miss.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"mix.v object *.n" )==1)) addnote2(sock,"replacement","place",rnode);
 if((len==1)&& (matchlike(vec,"mixed.v attribute *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"mixed.v attribute *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"moaned.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"moaned.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"mollify.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"motioned.v destination *.n" )==1)) addnote2(sock,"gesture","message",rnode);
 if((len==1)&& (matchlike(vec,"motions.v theme *.n" )==1)) addnote2(sock,"gesture","communicator",rnode);
 if((len==1)&& (matchlike(vec,"mounted.v object *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"mounted.v object *.n" )==1)) addnote2(sock,"placing","path",rnode);
 if((len==1)&& (matchlike(vec,"mourn.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"mourned.v method *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"mourning.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"mouthed.v theme *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"mouthing.v object *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"move.v coagent *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"move.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"mumble.v object *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"mumbled.v object *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"mumbled.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"mumbled.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"murdered.v attribute *.n" )==1)) addnote2(sock,"death","cause",rnode);
 if((len==1)&& (matchlike(vec,"murdered.v attribute *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"murdered.v theme *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"murmured.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"murmured.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"murmuring.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 /*
 if((len==1)&& (matchlike(vec,"must.v theme *.n" )==1)) addnote2(sock,"becoming-aware","ground",rnode);
 if((len==1)&& (matchlike(vec,"must.v theme *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"must.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"must.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"must.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 */
 if((len==1)&& (matchlike(vec,"muttered.v recipient *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"name.v object *.n" )==1)) addnote2(sock,"differentiation","phenomena",rnode);
 if((len==1)&& (matchlike(vec,"name.v theme *.n" )==1)) addnote2(sock,"differentiation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"nibbled.v attribute *.n" )==1)) addnote2(sock,"removing","path",rnode);
 if((len==1)&& (matchlike(vec,"nod.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"nod.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"nodded.v destination *.n" )==1)) addnote2(sock,"body-movement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"nodded.v destination *.n" )==1)) addnote2(sock,"gesture","addressee",rnode);
 if((len==1)&& (matchlike(vec,"nodded.v destination *.n" )==1)) addnote2(sock,"gesture","message",rnode);
 if((len==1)&& (matchlike(vec,"note.v theme *.n" )==1)) addnote2(sock,"becoming-aware","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"noted.v object *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"noted.v theme *.n" )==1)) addnote2(sock,"transportation","path",rnode);
 if((len==1)&& (matchlike(vec,"notice.v location *.n" )==1)) addnote2(sock,"becoming-aware","ground",rnode);
 if((len==1)&& (matchlike(vec,"notice.v location-time *.n" )==1)) addnote2(sock,"becoming-aware","ground",rnode);
 if((len==1)&& (matchlike(vec,"notice.v object *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"noticed.v location *.n" )==1)) addnote2(sock,"becoming-aware","ground",rnode);
 if((len==1)&& (matchlike(vec,"noticed.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"nudged.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"nudged.v theme *.n" )==1)) addnote2(sock,"manipulation","agent",rnode);
 if((len==1)&& (matchlike(vec,"nudges.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"nudges.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"nudging.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"nudging.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"nudging.v theme *.n" )==1)) addnote2(sock,"manipulation","agent",rnode);
 if((len==1)&& (matchlike(vec,"observe.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"observed.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"open.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"open.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"open.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"ordered.v theme *.n" )==1)) addnote2(sock,"request","addressee",rnode);
 if((len==1)&& (matchlike(vec,"ordered.v theme *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"overheard.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"oversees.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"paced.v destination *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"paced.v object *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"paced.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"pacing.v object *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"pack.v location *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"pack.v location *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"pack.v object *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"pack.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"packed.v coagent-insturment *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"packed.v coagent-insturment *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"packed.v location-time *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"packed.v location-time *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"packed.v object *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"packed.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"padded.v destination *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"padded.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"padded.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"padded.v object *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"padded.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"parachute.v location *.n" )==1)) addnote2(sock,"transportation","goal",rnode);
 if((len==1)&& (matchlike(vec,"parachute.v theme *.n" )==1)) addnote2(sock,"transportation","driver",rnode);
 if((len==1)&& (matchlike(vec,"parade.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"park.v location *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"park.v location-time *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"parked.v location *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"parked.v location-time *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"parked.v medium *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"parked.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"parking.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"partitioned.v method *.n" )==1)) addnote2(sock,"separation","criterion",rnode);
 if((len==1)&& (matchlike(vec,"partitioned.v theme *.n" )==1)) addnote2(sock,"separation","whole",rnode);
 if((len==1)&& (matchlike(vec,"passed.v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"passed.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"pattered.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"paused.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"pay.v object *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"pay.v object *.n" )==1)) addnote2(sock,"commerce","payment",rnode);
 if((len==1)&& (matchlike(vec,"paying.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"pealed.v location-time *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"peeked.v location-time *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"peeked.v location-time *.n" )==1)) addnote2(sock,"perception_active","manner",rnode);
 if((len==1)&& (matchlike(vec,"peeked.v medium *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"peeked.v medium *.n" )==1)) addnote2(sock,"perception_active","manner",rnode);
 if((len==1)&& (matchlike(vec,"peeking.v location *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"peep.v location *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"peep.v location-time *.n" )==1)) addnote2(sock,"perception-experience","direction",rnode);
 if((len==1)&& (matchlike(vec,"peep.v location-time *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"peep.v medium *.n" )==1)) addnote2(sock,"perception-experience","direction",rnode);
 if((len==1)&& (matchlike(vec,"peep.v medium *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"peep.v theme *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 if((len==1)&& (matchlike(vec,"peeped.v location-time *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"peer.v location *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"peered.v location *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"peered.v location-time *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==1)&& (matchlike(vec,"peered.v source *.n" )==1)) addnote2(sock,"perception_active","location of perceiver",rnode);
 if((len==1)&& (matchlike(vec,"peering.v source *.n" )==1)) addnote2(sock,"perception_active","location of perceiver",rnode);
 if((len==1)&& (matchlike(vec,"perceived.v location-time *.n" )==1)) addnote2(sock,"perception","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceived.v location-time *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceived.v medium *.n" )==1)) addnote2(sock,"perception","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceived.v medium *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceived.v object *.n" )==1)) addnote2(sock,"perception","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceived.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceived.v theme *.n" )==1)) addnote2(sock,"perception","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceived.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceives.v object *.n" )==1)) addnote2(sock,"perception","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceives.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"perceives.v theme *.n" )==1)) addnote2(sock,"perception-experience","perceiver-passive",rnode);
 if((len==1)&& (matchlike(vec,"perish.v method *.n" )==1)) addnote2(sock,"death","cause",rnode);
 if((len==1)&& (matchlike(vec,"perish.v method *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"perished.v source *.n" )==1)) addnote2(sock,"death","cause",rnode);
 if((len==1)&& (matchlike(vec,"perished.v theme *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"perishes.v theme *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"permitted.v recipient *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"perspiring.v attribute *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"persuade.v object *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"peruse.v object *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"phrased.v theme *.n" )==1)) addnote2(sock,"encoding","message",rnode);
 if((len==1)&& (matchlike(vec,"phrasing.v method *.n" )==1)) addnote2(sock,"encoding","manner",rnode);
 if((len==1)&& (matchlike(vec,"pick.v location *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"pick.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"pierced.v agent *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"pinched.v coagent-insturment *.n" )==1)) addnote2(sock,"manipulation","agent's body part",rnode);
 if((len==1)&& (matchlike(vec,"pinched.v location *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"pinched.v object *.n" )==1)) addnote2(sock,"manipulation","agent's body part",rnode);
 if((len==1)&& (matchlike(vec,"pinpointed.v theme *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"pitch.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"pity.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"placate.v object *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"place.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"plan.v object *.n" )==1)) addnote2(sock,"transportation","path",rnode);
 if((len==1)&& (matchlike(vec,"plan.v theme *.n" )==1)) addnote2(sock,"transportation","driver",rnode);
 if((len==1)&& (matchlike(vec,"planned.v theme *.n" )==1)) addnote2(sock,"questioning","speaker",rnode);
 if((len==1)&& (matchlike(vec,"planted.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"planted.v theme *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"pleased.v agent *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"pleased.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"pleased.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"pledge.v object *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"pledges.v theme *.n" )==1)) addnote2(sock,"commitment","speaker",rnode);
 if((len==1)&& (matchlike(vec,"plodded.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"plucked.v source *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"plummets.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"polished.v attribute *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"pondered.v theme *.n" )==1)) addnote2(sock,"cogitation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"position.v contact *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"position.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"possessed.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"pounced.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"pouted.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"praise.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"prattled.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"preached.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"preaches.v destination *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"preaching.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"predict.v object *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"predict.v theme *.n" )==1)) addnote2(sock,"expectation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"predicted.v location-time *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"predicted.v theme *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"prepared.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"presides.v theme *.n" )==1)) addnote2(sock,"leadership","leader",rnode);
 if((len==1)&& (matchlike(vec,"press.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"pressed.v object *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"presumed.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"presumed.v theme *.n" )==1)) addnote2(sock,"awareness","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"prevailed.v theme *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"prickled[!].v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"proclaimed.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"proclaims.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"promised.v destination *.n" )==1)) addnote2(sock,"commitment","addressee",rnode);
 if((len==1)&& (matchlike(vec,"promised.v destination *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"proposed.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"prove.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"prove.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"prowling.v location-time *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"prowling.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"pucker.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"puckered.v coagent *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"puckered.v coagent-insturment *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"puckered.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"punted.v destination *.n" )==1)) addnote2(sock,"transportation","goal",rnode);
 if((len==1)&& (matchlike(vec,"punted.v object *.n" )==1)) addnote2(sock,"transportation","distance",rnode);
 if((len==1)&& (matchlike(vec,"punting.v object *.n" )==1)) addnote2(sock,"transportation","cargo + passenger",rnode);
 if((len==1)&& (matchlike(vec,"purchased.v object *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"purchased.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"purge.v object *.n" )==1)) addnote2(sock,"emptying","theme",rnode);
 if((len==1)&& (matchlike(vec,"purge.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"purge.v source *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"purge.v source *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"purged.v theme *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"purged.v theme *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"purr.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"pursed.v attribute *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"pursed.v coagent-insturment *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"pursed.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"pursed.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"pursing.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"pursued.v object *.n" )==1)) addnote2(sock,"cotheme","cotheme",rnode);
 if((len==1)&& (matchlike(vec,"pursued.v theme *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"pushed.v contact *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"pushed.v destination *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"pushed.v method *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"pushed.v method *.n" )==1)) addnote2(sock,"manipulation","agent",rnode);
 if((len==1)&& (matchlike(vec,"pushed.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"pushed.v object *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"put.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"put.v object *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"put.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"puttered.v theme *.n" )==1)) addnote2(sock,"motion_noise","path",rnode);
 if((len==1)&& (matchlike(vec,"puttered.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"putting.v object *.n" )==1)) addnote2(sock,"categorization","category",rnode);
 if((len==1)&& (matchlike(vec,"puzzled.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"puzzled.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"quarrel.v coagent *.n" )==1)) addnote2(sock,"conversation","interlocutor-2",rnode);
 if((len==1)&& (matchlike(vec,"question.v object *.n" )==1)) addnote2(sock,"questioning","addressee",rnode);
 if((len==1)&& (matchlike(vec,"question.v theme *.n" )==1)) addnote2(sock,"questioning","speaker",rnode);
 if((len==1)&& (matchlike(vec,"questioned.v theme *.n" )==1)) addnote2(sock,"questioning","message",rnode);
 if((len==1)&& (matchlike(vec,"questioning.v object *.n" )==1)) addnote2(sock,"questioning","addressee",rnode);
 if((len==1)&& (matchlike(vec,"quit.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"quitted.v object *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"raised.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"raises.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"ran.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"ran.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"ran.v origin *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"ran.v recipient *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"ran.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"ran.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"ranted.v theme *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"ranted.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"rapped.v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"rapped.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"rasped.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"rattle.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"rattle.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"rattled.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"rattled.v method *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"reached.v destination *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"reached.v destination *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"reached.v location-time *.n" )==1)) addnote2(sock,"path-shape","area",rnode);
 if((len==1)&& (matchlike(vec,"reached.v object *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"reached.v object *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"reached.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"reaches.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"read.v location-time *.n" )==1)) addnote2(sock,"hear","medium",rnode);
 if((len==1)&& (matchlike(vec,"read.v medium *.n" )==1)) addnote2(sock,"hear","medium",rnode);
 if((len==1)&& (matchlike(vec,"read.v object *.n" )==1)) addnote2(sock,"hear","medium",rnode);
 if((len==1)&& (matchlike(vec,"read.v object *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"realised[!].v object *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"realize.v object *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"realized.v object *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"reasons.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"reassure.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"reassured.v method *.n" )==1)) addnote2(sock,"body-movement","cause",rnode);
 if((len==1)&& (matchlike(vec,"rebuked.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"rebuked.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"reclining.v attribute *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"recognized.v agent *.n" )==1)) addnote2(sock,"becoming-aware","evidence",rnode);
 if((len==1)&& (matchlike(vec,"recognized.v method *.n" )==1)) addnote2(sock,"becoming-aware","evidence",rnode);
 if((len==1)&& (matchlike(vec,"recount.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"recounting.v location-time *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"recounting.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"recovered.v theme *.n" )==1)) addnote2(sock,"recovery","patient",rnode);
 if((len==1)&& (matchlike(vec,"recovering.v theme *.n" )==1)) addnote2(sock,"recovery","patient",rnode);
 if((len==1)&& (matchlike(vec,"redefine.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"redefined.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"redefined.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"reduced.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"reformed.v attribute *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"regarded.v location *.n" )==1)) addnote2(sock,"categorization","category",rnode);
 if((len==1)&& (matchlike(vec,"regarded.v location-time *.n" )==1)) addnote2(sock,"categorization","category",rnode);
 if((len==1)&& (matchlike(vec,"regarded.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"regarded.v theme *.n" )==1)) addnote2(sock,"categorization","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"regarding.v object *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 if((len==1)&& (matchlike(vec,"registered.v attribute *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"registered.v theme *.n" )==1)) addnote2(sock,"becoming-aware","ground",rnode);
 if((len==1)&& (matchlike(vec,"regrets.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"regretted.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"regretted.v timeperiod *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"regretting.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"reiterates.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"rejoined.v theme *.n" )==1)) addnote2(sock,"communication_response","message",rnode);
 if((len==1)&& (matchlike(vec,"related.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"related.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"relating.v destination *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 if((len==1)&& (matchlike(vec,"relaxed.v attribute *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"relished.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"relished.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"relishes.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"remained.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"remember.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"remount.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"remove.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"removed.v object *.n" )==1)) addnote2(sock,"removing","goal",rnode);
 if((len==1)&& (matchlike(vec,"removing.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"renewed.v attribute *.n" )==1)) addnote2(sock,"appearance","inference",rnode);
 if((len==1)&& (matchlike(vec,"rented.v attribute *.n" )==1)) addnote2(sock,"motion_noise","path",rnode);
 if((len==1)&& (matchlike(vec,"rented.v theme *.n" )==1)) addnote2(sock,"commerce","buyer",rnode);
 if((len==1)&& (matchlike(vec,"rented.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"renting.v object *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"replace.v method *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"replace.v object *.n" )==1)) addnote2(sock,"replacement","theme-old",rnode);
 if((len==1)&& (matchlike(vec,"replaced.v method *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"replaced.v object *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"replaced.v theme *.n" )==1)) addnote2(sock,"replacement","theme-old",rnode);
 if((len==1)&& (matchlike(vec,"replicated.v location-time *.n" )==1)) addnote2(sock,"duplication","goal",rnode);
 if((len==1)&& (matchlike(vec,"replicated.v theme *.n" )==1)) addnote2(sock,"duplication","original",rnode);
 if((len==1)&& (matchlike(vec,"reply.v destination *.n" )==1)) addnote2(sock,"communication_response","trigger",rnode);
 if((len==1)&& (matchlike(vec,"reply.v theme *.n" )==1)) addnote2(sock,"communication_response","speaker",rnode);
 if((len==1)&& (matchlike(vec,"report.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"reported.v attribute *.n" )==1)) addnote2(sock,"evidence","support",rnode);
 if((len==1)&& (matchlike(vec,"reported.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"reported.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"reported.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"reports.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"represent.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"represents.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"reproduce.v object *.n" )==1)) addnote2(sock,"duplication","original",rnode);
 if((len==1)&& (matchlike(vec,"reproduce.v theme *.n" )==1)) addnote2(sock,"duplication","creator",rnode);
 if((len==1)&& (matchlike(vec,"reproduced.v location-time *.n" )==1)) addnote2(sock,"duplication","goal",rnode);
 if((len==1)&& (matchlike(vec,"reproduces.v theme *.n" )==1)) addnote2(sock,"duplication","creator",rnode);
 if((len==1)&& (matchlike(vec,"reproved.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"request.v object *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"request.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"request.v theme *.n" )==1)) addnote2(sock,"hear","speaker",rnode);
 if((len==1)&& (matchlike(vec,"request.v theme *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"requested.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"rescue.v object *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"resent.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"resent.v object *.n" )==1)) addnote2(sock,"experiencer-subj","reason",rnode);
 if((len==1)&& (matchlike(vec,"resented.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"resented.v medium *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"resented.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"restrict.v object *.n" )==1)) addnote2(sock,"scrutiny","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"retailed.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"retreated.v location *.n" )==1)) addnote2(sock,"cotheme","goal",rnode);
 if((len==1)&& (matchlike(vec,"retreated.v location *.n" )==1)) addnote2(sock,"departing","goal",rnode);
 if((len==1)&& (matchlike(vec,"retreating.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"returning.v destination *.n" )==1)) addnote2(sock,"hear","topic",rnode);
 if((len==1)&& (matchlike(vec,"returns.v theme *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"returns.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"reveal.v destination *.n" )==1)) addnote2(sock,"evidence","proposition",rnode);
 if((len==1)&& (matchlike(vec,"reveal.v object *.n" )==1)) addnote2(sock,"evidence","proposition",rnode);
 if((len==1)&& (matchlike(vec,"reveal.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"revealed.v location-time *.n" )==1)) addnote2(sock,"evidence","proposition",rnode);
 if((len==1)&& (matchlike(vec,"reveals.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"reveals.v theme *.n" )==1)) addnote2(sock,"awareness","evidence",rnode);
 if((len==1)&& (matchlike(vec,"reveals.v theme *.n" )==1)) addnote2(sock,"evidence","support",rnode);
 if((len==1)&& (matchlike(vec,"ride.v object *.n" )==1)) addnote2(sock,"transportation","path",rnode);
 if((len==1)&& (matchlike(vec,"ride.v object *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 if((len==1)&& (matchlike(vec,"ride.v source *.n" )==1)) addnote2(sock,"transportation","source",rnode);
 if((len==1)&& (matchlike(vec,"riled.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"riling.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"ringing.v object *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"ringing.v object *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 if((len==1)&& (matchlike(vec,"risen.v source *.n" )==1)) addnote2(sock,"path-shape","source",rnode);
 if((len==1)&& (matchlike(vec,"risen.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"rises.v location *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"rises.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"rises.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"roam.v object *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"roam.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"roared.v location *.n" )==1)) addnote2(sock,"communication_noise","addressee",rnode);
 if((len==1)&& (matchlike(vec,"roared.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"roared.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"rock.v object *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"rode.v object *.n" )==1)) addnote2(sock,"transportation","path",rnode);
 if((len==1)&& (matchlike(vec,"rolled.v object *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"rolled.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"rolled.v object *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"romped.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"rose.v recipient *.n" )==1)) addnote2(sock,"path-shape","area",rnode);
 if((len==1)&& (matchlike(vec,"round.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"rouse.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"rowed.v destination *.n" )==1)) addnote2(sock,"arriving","goal",rnode);
 if((len==1)&& (matchlike(vec,"rowed.v destination *.n" )==1)) addnote2(sock,"transportation","goal",rnode);
 if((len==1)&& (matchlike(vec,"rowed.v location *.n" )==1)) addnote2(sock,"transportation","goal",rnode);
 if((len==1)&& (matchlike(vec,"rowed.v manner *.n" )==1)) addnote2(sock,"transportation","distance",rnode);
 if((len==1)&& (matchlike(vec,"rowed.v object *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 if((len==1)&& (matchlike(vec,"rowing.v theme *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 if((len==1)&& (matchlike(vec,"rows.v object *.n" )==1)) addnote2(sock,"transportation","distance",rnode);
 if((len==1)&& (matchlike(vec,"rows.v object *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 if((len==1)&& (matchlike(vec,"rue.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"ruled.v theme *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"rules.v location-time *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"rumbling.v attribute *.n" )==1)) addnote2(sock,"adornment","theme",rnode);
 if((len==1)&& (matchlike(vec,"rumbling.v attribute *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"ruminate.v theme *.n" )==1)) addnote2(sock,"cogitation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"run.v attribute *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"run.v manner *.n" )==1)) addnote2(sock,"self-motion","distance",rnode);
 if((len==1)&& (matchlike(vec,"run.v object *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"run.v recipient *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"run.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"run.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"run.v theme *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"run.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"running.v object *.n" )==1)) addnote2(sock,"self-motion","distance",rnode);
 if((len==1)&& (matchlike(vec,"runs.v object *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"rush.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"rush.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"rushed.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"rushing.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"rustled.v location-time *.n" )==1)) addnote2(sock,"motion_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"rustled.v location-time *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"rustled.v medium *.n" )==1)) addnote2(sock,"motion_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"rustled.v medium *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"rustled.v theme *.n" )==1)) addnote2(sock,"perception_noise","causer",rnode);
 if((len==1)&& (matchlike(vec,"rustles.v theme *.n" )==1)) addnote2(sock,"motion_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"rustles.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"rustling.v location-time *.n" )==1)) addnote2(sock,"motion_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"rustling.v location-time *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"saddened.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"saddened.v theme *.n" )==1)) addnote2(sock,"death","cause",rnode);
 if((len==1)&& (matchlike(vec,"said.v theme *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"said.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"sailed.v location-time *.n" )==1)) addnote2(sock,"transportation","area",rnode);
 if((len==1)&& (matchlike(vec,"sailed.v medium *.n" )==1)) addnote2(sock,"transportation","area",rnode);
 if((len==1)&& (matchlike(vec,"sailed.v theme *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 if((len==1)&& (matchlike(vec,"sang.v object *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"sang.v object *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"sang.v theme *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"sank.v location *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"sat.v coagent-insturment *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"sat.v theme *.n" )==1)) addnote2(sock,"conversation","interlocutors",rnode);
 if((len==1)&& (matchlike(vec,"saunter.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"sauntered.v destination *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"sauntered.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"sauntering.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"sauntering.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"savouring.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"saw.v object *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"saw.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"saw.v theme *.n" )==1)) addnote2(sock,"transportation","driver",rnode);
 if((len==1)&& (matchlike(vec,"say.v object *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"say.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"say.v object *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"say.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"scampered.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"scampered.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"scanned.v object *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"scanned.v theme *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"scans.v object *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"scared.v attribute *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"scolded.v location-time *.n" )==1)) addnote2(sock,"judgment","expressor",rnode);
 if((len==1)&& (matchlike(vec,"scolded.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"scowled.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"screamed.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"screeched.v location *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"screeched.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"screeches.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"scrutinized.v object *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"scuttle.v object *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"scuttle.v object *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"scuttle.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"scuttle.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"scuttled.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"search.v object *.n" )==1)) addnote2(sock,"appearance","ground",rnode);
 if((len==1)&& (matchlike(vec,"searched.v medium *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"searched.v object *.n" )==1)) addnote2(sock,"appearance","ground",rnode);
 if((len==1)&& (matchlike(vec,"searched.v object *.n" )==1)) addnote2(sock,"perception-experience","ground",rnode);
 if((len==1)&& (matchlike(vec,"searched.v object *.n" )==1)) addnote2(sock,"perception_active","ground",rnode);
 if((len==1)&& (matchlike(vec,"searched.v theme *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"searching.v location-time *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"see.v manner *.n" )==1)) addnote2(sock,"perception-experience","direction",rnode);
 if((len==1)&& (matchlike(vec,"see.v object *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"see.v object *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"see.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"see.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"see.v source *.n" )==1)) addnote2(sock,"perception-experience","location of perceiver",rnode);
 if((len==1)&& (matchlike(vec,"seeing.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"seem.v theme *.n" )==1)) addnote2(sock,"coming-to-believe","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"seemed.v analogy *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"seemed.v theme *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"seemed.v theme *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"seemed.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"seems.v analogy *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"seems.v analogy *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"seems.v theme *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"seen.v object *.n" )==1)) addnote2(sock,"cotheme","cotheme",rnode);
 if((len==1)&& (matchlike(vec,"sees.v method *.n" )==1)) addnote2(sock,"perception-experience","ground",rnode);
 if((len==1)&& (matchlike(vec,"select.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"sell.v object *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"selling.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"send.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"sense.v location-time *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"sense.v object *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"sensed.v location *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"sensed.v location-time *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"sensed.v object *.n" )==1)) addnote2(sock,"perception-experience","perceiver-passive",rnode);
 if((len==1)&& (matchlike(vec,"senses.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"sent.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"set.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"shadowed.v agent *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"shadowed.v method *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"shadowed.v theme *.n" )==1)) addnote2(sock,"cotheme","cotheme",rnode);
 if((len==1)&& (matchlike(vec,"shaken.v method *.n" )==1)) addnote2(sock,"body-movement","cause",rnode);
 if((len==1)&& (matchlike(vec,"shaken.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"shaped.v attribute *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"shattering.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"shed.v attribute *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"shifted.v object *.n" )==1)) addnote2(sock,"motion_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"shifted.v object *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"shock.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"shocked.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"shocked.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"shocked.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"shook.v object *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"shook.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"shook.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"shook.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"shook.v theme *.n" )==1)) addnote2(sock,"manipulation","agent",rnode);
 if((len==1)&& (matchlike(vec,"shook.v theme *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"shook.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"shot.v attribute *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"shot.v object *.n" )==1)) addnote2(sock,"judgment","expressor",rnode);
 
 /*
 if((len==1)&& (matchlike(vec,"should.v theme *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"should.v theme *.n" )==1)) addnote2(sock,"removing","agent",rnode);
 if((len==1)&& (matchlike(vec,"should.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 */
 if((len==1)&& (matchlike(vec,"shouted.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"shoved.v object *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"shoved.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"shrieked.v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"shrieks.v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"shrink.v location-time *.n" )==1)) addnote2(sock,"expansion","dimension",rnode);
 if((len==1)&& (matchlike(vec,"shrink.v medium *.n" )==1)) addnote2(sock,"expansion","dimension",rnode);
 if((len==1)&& (matchlike(vec,"shrinking.v theme *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"shrinks.v object *.n" )==1)) addnote2(sock,"expansion","item",rnode);
 if((len==1)&& (matchlike(vec,"shrugged.v location *.n" )==1)) addnote2(sock,"body-movement","cause",rnode);
 if((len==1)&& (matchlike(vec,"shrugged.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"shrugged.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"shuddered.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"shuffled.v destination *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"shuffled.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"shuffled.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"shuffled.v source *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"shuffled.v source *.n" )==1)) addnote2(sock,"self-motion","manner",rnode);
 if((len==1)&& (matchlike(vec,"shuffling.v location *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"shuffling.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"shut.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"signalled.v object *.n" )==1)) addnote2(sock,"gesture","addressee",rnode);
 if((len==1)&& (matchlike(vec,"signed.v attribute *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"silhouetted.v attribute *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"simmered.v theme *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"simulated.v attribute *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"simulated.v attribute *.n" )==1)) addnote2(sock,"transportation","goal",rnode);
 if((len==1)&& (matchlike(vec,"sing.v object *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"sing.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"single.v object *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"sings.v destination *.n" )==1)) addnote2(sock,"manner","addressee",rnode);
 if((len==1)&& (matchlike(vec,"sings.v object *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"sizzled.v location-time *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"sizzled.v medium *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"skate.v destination *.n" )==1)) addnote2(sock,"transportation","goal",rnode);
 if((len==1)&& (matchlike(vec,"skim.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"skipped.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"skipped.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"skipped.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"skips.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"skips.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"skirted.v object *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"slanted.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"slants.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"slap.v object *.n" )==1)) addnote2(sock,"heat","cause",rnode);
 if((len==1)&& (matchlike(vec,"slap.v object *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"slapped.v coagent-insturment *.n" )==1)) addnote2(sock,"perception_noise","manner",rnode);
 if((len==1)&& (matchlike(vec,"slaughter.v object *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"sleeping.v attribute *.n" )==1)) addnote2(sock,"departing","manner",rnode);
 if((len==1)&& (matchlike(vec,"sleeping.v attribute *.n" )==1)) addnote2(sock,"perception-experience","manner",rnode);
 if((len==1)&& (matchlike(vec,"slink.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"slink.v location-time *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"slink.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"slinking.v location *.n" )==1)) addnote2(sock,"arriving","goal",rnode);
 if((len==1)&& (matchlike(vec,"slinking.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"slithered.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"slithered.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"slurring.v object *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"smarting.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"smear.v object *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"smear.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"smear.v theme *.n" )==1)) addnote2(sock,"filling","agent",rnode);
 if((len==1)&& (matchlike(vec,"smear.v theme *.n" )==1)) addnote2(sock,"placing","agent",rnode);
 if((len==1)&& (matchlike(vec,"smeared.v coagent *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"smeared.v coagent *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"smeared.v coagent-insturment *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"smeared.v coagent-insturment *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"smeared.v object *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"smeared.v object *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"smelled.v coagent-insturment *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"smelled.v coagent-insturment *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"smelled.v coagent-insturment *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"smells.v analogy *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"smells.v source *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"smells.v source *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"smells.v source *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"smelt.v object *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"smelt.v object *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"smelt.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"smelt.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"smelt.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"smelt.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"smelt.v theme *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"smiled.v coagent *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"smiled.v location *.n" )==1)) addnote2(sock,"body-movement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"smiled.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"smiled.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"smiles.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"smiling.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"smirked.v coagent-insturment *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"smirked.v location *.n" )==1)) addnote2(sock,"body-movement","cause",rnode);
 if((len==1)&& (matchlike(vec,"smirked.v location-time *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"smirked.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"smoked.v attribute *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"smouldered.v coagent-insturment *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"smouldered.v location-time *.n" )==1)) addnote2(sock,"heat","location",rnode);
 if((len==1)&& (matchlike(vec,"smouldered.v theme *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"smouldered.v theme *.n" )==1)) addnote2(sock,"heat","location",rnode);
 if((len==1)&& (matchlike(vec,"smouldering.v location-time *.n" )==1)) addnote2(sock,"heat","location",rnode);
 if((len==1)&& (matchlike(vec,"smouldering.v theme *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"smouldering.v theme *.n" )==1)) addnote2(sock,"heat","location",rnode);
 if((len==1)&& (matchlike(vec,"snaked.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"snarled.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"snarled.v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"snatched.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"sneaked.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"sneaked.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"sneaked.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"sniff.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"snorted.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"soaked.v attribute *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"sobbed.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"solved.v object *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"soothe.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"sorting.v location *.n" )==1)) addnote2(sock,"differentiation","phenomenon-2",rnode);
 if((len==1)&& (matchlike(vec,"sorting.v object *.n" )==1)) addnote2(sock,"differentiation","phenomena",rnode);
 if((len==1)&& (matchlike(vec,"sorting.v object *.n" )==1)) addnote2(sock,"differentiation","phenomenon-1",rnode);
 if((len==1)&& (matchlike(vec,"sound.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"sounded.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"sounding.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"sounds.v source *.n" )==1)) addnote2(sock,"appearance","ground",rnode);
 if((len==1)&& (matchlike(vec,"speak.v location *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"speak.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"splintered.v attribute *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"spoke.v location-time *.n" )==1)) addnote2(sock,"conversation","medium",rnode);
 if((len==1)&& (matchlike(vec,"spoken.v method *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"spoken.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"spotted.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"sprang.v recipient *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"sprang.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"sprawled.v medium *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"sprawled.v theme *.n" )==1)) addnote2(sock,"posture","agent",rnode);
 if((len==1)&& (matchlike(vec,"spread.v object *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"spread.v object *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"spread.v object *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"spread.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"spread.v theme *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"spread.v theme *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"spring.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"sprinting.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"sprints.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"sprints.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"squeak.v location-time *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"squeak.v medium *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"squeaked.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"squealed.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"squeeze.v location-time *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"squeeze.v medium *.n" )==1)) addnote2(sock,"manipulation","entity",rnode);
 if((len==1)&& (matchlike(vec,"squeezed.v agent *.n" )==1)) addnote2(sock,"manipulation","agent",rnode);
 if((len==1)&& (matchlike(vec,"squeezed.v coagent-insturment *.n" )==1)) addnote2(sock,"manipulation","agent's body part",rnode);
 if((len==1)&& (matchlike(vec,"squeezed.v theme *.n" )==1)) addnote2(sock,"manipulation","agent's body part",rnode);
 if((len==1)&& (matchlike(vec,"stagger.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"staggered.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"staggered.v source *.n" )==1)) addnote2(sock,"perception_noise","source",rnode);
 if((len==1)&& (matchlike(vec,"staggered.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"stamped.v contact *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"stamped.v destination *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"stamped.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"stamped.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"stamped.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"stand.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"stands.v theme *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"stared.v object *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"stared.v source *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"stared.v theme *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 if((len==1)&& (matchlike(vec,"stares.v object *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 if((len==1)&& (matchlike(vec,"staring.v location-time *.n" )==1)) addnote2(sock,"perception_active","manner",rnode);
 if((len==1)&& (matchlike(vec,"staring.v location-time *.n" )==1)) addnote2(sock,"placing","manner",rnode);
 if((len==1)&& (matchlike(vec,"starring.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"starring.v theme *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"start.v theme *.n" )==1)) addnote2(sock,"duplication","creator",rnode);
 if((len==1)&& (matchlike(vec,"started.v object *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"started.v object *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"started.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"started.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"starting.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"startled.v attribute *.n" )==1)) addnote2(sock,"self-motion","manner",rnode);
 if((len==1)&& (matchlike(vec,"startled.v object *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"startled.v theme *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"state.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"statesman[?].v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"stay.v theme *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"steal.v object *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"stealing.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"steamed.v attribute *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"steamed.v attribute *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"steamed.v attribute *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"step.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"step.v manner *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"stepped.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"steps.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"stew.v location-time *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"stew.v medium *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"stimulated.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"stimulated.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"sting.v object *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"sting.v object *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"sting.v object *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"stirred.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"stirred.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"stole.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"stole.v object *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"stole.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"stood.v theme *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 if((len==1)&& (matchlike(vec,"stooped.v theme *.n" )==1)) addnote2(sock,"posture","agent",rnode);
 if((len==1)&& (matchlike(vec,"stop.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"stowed.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"stretching.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"strike.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"strip.v object *.n" )==1)) addnote2(sock,"emptying","theme",rnode);
 if((len==1)&& (matchlike(vec,"strip.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"stripped.v source *.n" )==1)) addnote2(sock,"emptying","source",rnode);
 if((len==1)&& (matchlike(vec,"stripped.v source *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"strolls.v location-time *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"strutted.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"stuck.v location-time *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"stuck.v location-time *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"stuck.v medium *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"stuck.v medium *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"stuff.v object *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"stuff.v object *.n" )==1)) addnote2(sock,"placing","goal",rnode);
 if((len==1)&& (matchlike(vec,"stumbled.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"stumbling.v location-time *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"stumbling.v medium *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"stung.v agent *.n" )==1)) addnote2(sock,"body","agent",rnode);
 if((len==1)&& (matchlike(vec,"stung.v method *.n" )==1)) addnote2(sock,"body","agent",rnode);
 if((len==1)&& (matchlike(vec,"stung.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"stunned.v attribute *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"stunned.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"stunned.v method *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"substituted.v method *.n" )==1)) addnote2(sock,"replacement","agent",rnode);
 if((len==1)&& (matchlike(vec,"substituted.v theme *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"substituted.v theme *.n" )==1)) addnote2(sock,"replacement","theme-old",rnode);
 if((len==1)&& (matchlike(vec,"succeed.v object *.n" )==1)) addnote2(sock,"replacement","theme-old",rnode);
 if((len==1)&& (matchlike(vec,"suggest.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"suggest.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"summoned.v object *.n" )==1)) addnote2(sock,"request","addressee",rnode);
 if((len==1)&& (matchlike(vec,"summons.v object *.n" )==1)) addnote2(sock,"request","addressee",rnode);
 if((len==1)&& (matchlike(vec,"supplant.v theme *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"supplanted.v agent *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"supplanted.v agent *.n" )==1)) addnote2(sock,"replacement","theme-old",rnode);
 if((len==1)&& (matchlike(vec,"surprised.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"surprised.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"suspect.v location-time *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"suspect.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"swaggered.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"swaggered.v recipient *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"swam.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"swam.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"swam.v theme *.n" )==1)) addnote2(sock,"cotheme","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"swap.v coagent-insturment *.n" )==1)) addnote2(sock,"replacement","theme-new",rnode);
 if((len==1)&& (matchlike(vec,"swap.v theme *.n" )==1)) addnote2(sock,"replacement","theme-old",rnode);
 if((len==1)&& (matchlike(vec,"swapped.v theme *.n" )==1)) addnote2(sock,"replacement","themes",rnode);
 if((len==1)&& (matchlike(vec,"swapping.v object *.n" )==1)) addnote2(sock,"replacement","themes",rnode);
 if((len==1)&& (matchlike(vec,"swayed.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"swerved.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"swim.v location *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"swim.v location-time *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"swimming.v attribute *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"swimming.v location-time *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"swing.v destination *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"swing.v destination *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"swing.v manner *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"swing.v manner *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"swing.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"swing.v object *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"swiping.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"swished.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"switch.v theme *.n" )==1)) addnote2(sock,"replacement","agent",rnode);
 if((len==1)&& (matchlike(vec,"switched.v destination *.n" )==1)) addnote2(sock,"replacement","place-2",rnode);
 if((len==1)&& (matchlike(vec,"swooped.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"sworn.v attribute *.n" )==1)) addnote2(sock,"commitment","speaker",rnode);
 if((len==1)&& (matchlike(vec,"sworn.v attribute *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"swung.v destination *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"swung.v destination *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"swung.v location-time *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"swung.v location-time *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"swung.v object *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"swung.v object *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"swung.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"symbolize.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"symbolizes.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"symbolizes.v theme *.n" )==1)) addnote2(sock,"categorization","category",rnode);
 if((len==1)&& (matchlike(vec,"sympathized.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"tailing.v object *.n" )==1)) addnote2(sock,"cotheme","cotheme",rnode);
 if((len==1)&& (matchlike(vec,"take.v destination *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"take.v object *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"take.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"take.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"take.v object *.n" )==1)) addnote2(sock,"judgment","expressor",rnode);
 if((len==1)&& (matchlike(vec,"take.v theme *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"take.v theme *.n" )==1)) addnote2(sock,"cure","patient",rnode);
 if((len==1)&& (matchlike(vec,"take.v theme *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"taken.v location *.n" )==1)) addnote2(sock,"removing","goal",rnode);
 if((len==1)&& (matchlike(vec,"taken.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"taking.v destination *.n" )==1)) addnote2(sock,"removing","goal",rnode);
 if((len==1)&& (matchlike(vec,"taking.v object *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"taking.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"talk.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"talked.v theme *.n" )==1)) addnote2(sock,"conversation","interlocutors",rnode);
 if((len==1)&& (matchlike(vec,"talked.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"taste.v object *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"taste.v object *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"taste.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"tastes.v theme *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"taxied.v theme *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 if((len==1)&& (matchlike(vec,"teach.v location-time *.n" )==1)) addnote2(sock,"transportation","area",rnode);
 if((len==1)&& (matchlike(vec,"telephoned.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"tell.v object *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"tell.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"tell.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"tells.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"tells.v object *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"tells.v theme *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"tells.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"tend.v theme *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"thank.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"thinking.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"thought.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"threatened.v theme *.n" )==1)) addnote2(sock,"commitment","addressee",rnode);
 if((len==1)&& (matchlike(vec,"threatened.v theme *.n" )==1)) addnote2(sock,"commitment","speaker",rnode);
 if((len==1)&& (matchlike(vec,"threw.v location *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"threw.v location *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"threw.v location-time *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"threw.v location-time *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"threw.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"threw.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"threw.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"throw.v theme *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"thrown.v agent *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"thrown.v agent *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"thrown.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"thrown.v theme *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"thrown.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"throws.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"thud.v location *.n" )==1)) addnote2(sock,"motion_noise","goal",rnode);
 if((len==1)&& (matchlike(vec,"thud.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"thudded.v location *.n" )==1)) addnote2(sock,"motion_noise","goal",rnode);
 if((len==1)&& (matchlike(vec,"thudding.v location *.n" )==1)) addnote2(sock,"motion_noise","goal",rnode);
 if((len==1)&& (matchlike(vec,"thumped.v theme *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"thunder.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"thundered.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"thundered.v object *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"thundering.v location *.n" )==1)) addnote2(sock,"communication_noise","addressee",rnode);
 if((len==1)&& (matchlike(vec,"tickle.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"tickle.v theme *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"tickles.v location *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"tickles.v theme *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"tickling.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"tiled.v attribute *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"tiled.v attribute *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"tiled.v attribute *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"till.v location-time *.n" )==1)) addnote2(sock,"path-shape","area",rnode);
 if((len==1)&& (matchlike(vec,"tingle.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"tingled.v theme *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"tingled.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"tingles.v coagent *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"tingles.v theme *.n" )==1)) addnote2(sock,"body","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"tingling.v attribute *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"tingling.v attribute *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"tingling.v coagent-insturment *.n" )==1)) addnote2(sock,"body","cause",rnode);
 if((len==1)&& (matchlike(vec,"tingling.v theme *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"tingling.v theme *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"tingling.v theme *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"tinkled.v location-time *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"tinkled.v theme *.n" )==1)) addnote2(sock,"perception_noise","causer",rnode);
 if((len==1)&& (matchlike(vec,"tinkled.v theme *.n" )==1)) addnote2(sock,"perception_noise","sound",rnode);
 if((len==1)&& (matchlike(vec,"tipped.v object *.n" )==1)) addnote2(sock,"motion_noise","path",rnode);
 if((len==1)&& (matchlike(vec,"tipped.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"tiptoe.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"tiptoed.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"tiptoed.v recipient *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"tired.v attribute *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"toddled.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"took.v source *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"toss.v location *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"toss.v location *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"tossed.v recipient *.n" )==1)) addnote2(sock,"body-movement","goal",rnode);
 if((len==1)&& (matchlike(vec,"tossed.v recipient *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"tossed.v source *.n" )==1)) addnote2(sock,"body-movement","source",rnode);
 if((len==1)&& (matchlike(vec,"tossed.v source *.n" )==1)) addnote2(sock,"cause-to-move","source",rnode);
 if((len==1)&& (matchlike(vec,"tossed.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"tossed.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"total.v object *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"tottered.v destination *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"tottered.v recipient *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"tottered.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"totters.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"touched.v object *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"touched.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"tracked.v object *.n" )==1)) addnote2(sock,"cotheme","cotheme",rnode);
 if((len==1)&& (matchlike(vec,"trade.v object *.n" )==1)) addnote2(sock,"replacement","places",rnode);
 if((len==1)&& (matchlike(vec,"trade.v object *.n" )==1)) addnote2(sock,"replacement","themes",rnode);
 if((len==1)&& (matchlike(vec,"trading.v object *.n" )==1)) addnote2(sock,"replacement","theme-old",rnode);
 if((len==1)&& (matchlike(vec,"trading.v object *.n" )==1)) addnote2(sock,"replacement","themes",rnode);
 if((len==1)&& (matchlike(vec,"translates.v object *.n" )==1)) addnote2(sock,"categorization","category",rnode);
 if((len==1)&& (matchlike(vec,"traps.v object *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"traps.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"treated.v coagent *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"trifle.v coagent-insturment *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"tripped.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"tripped.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"trot.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"trotted.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"trotted.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"trundled.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"trusted.v coagent-insturment *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"tucked.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"tug.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"tugged.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"tugged.v source *.n" )==1)) addnote2(sock,"cause-to-move","source",rnode);
 if((len==1)&& (matchlike(vec,"tugging.v location-time *.n" )==1)) addnote2(sock,"cause-to-move","path",rnode);
 if((len==1)&& (matchlike(vec,"tugging.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"turn.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"turned.v location *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"turned.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"turned.v source *.n" )==1)) addnote2(sock,"self-motion","source",rnode);
 if((len==1)&& (matchlike(vec,"turned.v theme *.n" )==1)) addnote2(sock,"request","addressee",rnode);
 if((len==1)&& (matchlike(vec,"turned.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"turned.v theme *.n" )==1)) addnote2(sock,"statement","addressee",rnode);
 if((len==1)&& (matchlike(vec,"turns.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"unconsidered[!].v attribute *.n" )==1)) addnote2(sock,"coming-to-believe","content",rnode);
 if((len==1)&& (matchlike(vec,"understand.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"understood.v method *.n" )==1)) addnote2(sock,"awareness","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"understood.v theme *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"unfixed[!].v attribute *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"unglazed[!].v attribute *.n" )==1)) addnote2(sock,"motion_noise","path",rnode);
 if((len==1)&& (matchlike(vec,"unlocked.v attribute *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"unlocked.v attribute *.n" )==1)) addnote2(sock,"path-shape","path",rnode);
 if((len==1)&& (matchlike(vec,"unlocked.v attribute *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"unlocked.v attribute *.n" )==1)) addnote2(sock,"scrutiny","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"unnerved.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"unsettle.v object *.n" )==1)) addnote2(sock,"cogitation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"unsettle.v object *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"unsettle.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"unsettled.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"unsettled.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"unsettled.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"upset.v attribute *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"upset.v method *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"upset.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"urge.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"urged.v object *.n" )==1)) addnote2(sock,"request","addressee",rnode);
 if((len==1)&& (matchlike(vec,"urged.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"urged.v theme *.n" )==1)) addnote2(sock,"request","speaker",rnode);
 if((len==1)&& (matchlike(vec,"urges.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"use.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"use.v object *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"use.v theme *.n" )==1)) addnote2(sock,"filling","goal",rnode);
 if((len==1)&& (matchlike(vec,"used.v theme *.n" )==1)) addnote2(sock,"appearance","characterization",rnode);
 if((len==1)&& (matchlike(vec,"vacate.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"vacated.v agent *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"vacated.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"vanished.v destination *.n" )==1)) addnote2(sock,"departing","goal",rnode);
 if((len==1)&& (matchlike(vec,"vanished.v origin *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"vanished.v source *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"vanished.v source *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"vanished.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"vanishing.v location-time *.n" )==1)) addnote2(sock,"departing","path",rnode);
 if((len==1)&& (matchlike(vec,"veered.v location *.n" )==1)) addnote2(sock,"path-shape","goal",rnode);
 if((len==1)&& (matchlike(vec,"veered.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"view.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"viewed.v coagent-insturment *.n" )==1)) addnote2(sock,"perception_active","manner",rnode);
 if((len==1)&& (matchlike(vec,"viewed.v theme *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 if((len==1)&& (matchlike(vec,"visit.v object *.n" )==1)) addnote2(sock,"arriving","goal",rnode);
 if((len==1)&& (matchlike(vec,"visit.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"visited.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"voice.v object *.n" )==1)) addnote2(sock,"encoding","message",rnode);
 if((len==1)&& (matchlike(vec,"voiced.v location-time *.n" )==1)) addnote2(sock,"encoding","message",rnode);
 if((len==1)&& (matchlike(vec,"voiced.v timeperiod *.n" )==1)) addnote2(sock,"encoding","message",rnode);
 if((len==1)&& (matchlike(vec,"voicing.v object *.n" )==1)) addnote2(sock,"encoding","message",rnode);
 if((len==1)&& (matchlike(vec,"volunteered.v theme *.n" )==1)) addnote2(sock,"commitment","message",rnode);
 if((len==1)&& (matchlike(vec,"waddled.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"wade.v location *.n" )==1)) addnote2(sock,"self-motion","goal",rnode);
 if((len==1)&& (matchlike(vec,"wade.v object *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"wagged.v object *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"wailed.v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"wailed.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"walking.v attribute *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"wandered.v method *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"want.v agent *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"want.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"want.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"want.v theme *.n" )==1)) addnote2(sock,"self-motion","area",rnode);
 if((len==1)&& (matchlike(vec,"wanted.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"wanted.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"wants.v location-time *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"wants.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"warbled.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 /*
 if((len==1)&& (matchlike(vec,"was.v location-time *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"cogitation","topic",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"communication_response","message",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"judgment","evaluee",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"motion_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"path-shape","distance",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"was.v object *.n" )==1)) addnote2(sock,"transportation","distance",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"appearance","body part",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"cotheme","cotheme",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"differentiation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"duplication","goal",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"duplication","original",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"invention","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"manipulation","agent",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"manner","speaker",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"motion","path",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"path-shape","road",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"perception","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"perception-experience","body part",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"perception-experience","manner",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"perception_active","perceiver-agentive",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"recovery","patient",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"self-motion","path",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"separation","whole",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"statement","message",rnode);
 if((len==1)&& (matchlike(vec,"was.v theme *.n" )==1)) addnote2(sock,"statement","speaker",rnode);
 if((len==1)&& (matchlike(vec,"was.v timeperiod *.n" )==1)) addnote2(sock,"statement","message",rnode);
 */
   if((len==1)&& (matchlike(vec,"waste.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"watched.v object *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==1)&& (matchlike(vec,"waved.v location *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"waved.v location *.n" )==1)) addnote2(sock,"gesture","message",rnode);
 if((len==1)&& (matchlike(vec,"waves.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"weeping.v attribute *.n" )==1)) addnote2(sock,"manipulation","agent",rnode);
 if((len==1)&& (matchlike(vec,"weeping.v attribute *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 /*
 if((len==1)&& (matchlike(vec,"were.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 if((len==1)&& (matchlike(vec,"were.v object *.n" )==1)) addnote2(sock,"body","body part",rnode);
 if((len==1)&& (matchlike(vec,"were.v object *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"were.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"were.v object *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"were.v object *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"becoming-aware","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"categorization","item",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"commitment","speaker",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"death","protagonist",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"differentiation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"expectation","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"heat","emotion",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"heat","location",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"imitation","copy",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"joining","parts",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"judgment","reason",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"leadership","jurisdiction",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"perception-experience","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"perception_active","body part",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"perception_noise","cause",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"request","addressee",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"request","message",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"scrutiny","ground",rnode);
 if((len==1)&& (matchlike(vec,"were.v theme *.n" )==1)) addnote2(sock,"transportation","vehicle",rnode);
 */
  if((len==1)&& (matchlike(vec,"whined.v object *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"whined.v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"whirled.v theme *.n" )==1)) addnote2(sock,"body-movement","cause",rnode);
 if((len==1)&& (matchlike(vec,"whirred.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"whirring.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"whispered.v theme *.n" )==1)) addnote2(sock,"hear","message",rnode);
 if((len==1)&& (matchlike(vec,"whistle.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"whooped.v theme *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"wiggle.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"wiggle.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"wiggled.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"wiggled.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"wiggles.v object *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 /*
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"cogitation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"commerce","goods",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"cure","affliction",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"duplication","creator",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"expectation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"recovery","affliction",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"recovery","body part",rnode);
 if((len==1)&& (matchlike(vec,"will.v theme *.n" )==1)) addnote2(sock,"replacement","theme-old",rnode);
 */
 if((len==1)&& (matchlike(vec,"wink.v theme *.n" )==1)) addnote2(sock,"body-movement","agent",rnode);
 if((len==1)&& (matchlike(vec,"winked.v location-time *.n" )==1)) addnote2(sock,"body-movement","path",rnode);
 if((len==1)&& (matchlike(vec,"winked.v location-time *.n" )==1)) addnote2(sock,"self-motion","manner",rnode);
 if((len==1)&& (matchlike(vec,"withdraw.v object *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"withdraw.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"withdrawing.v object *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"withdrawing.v object *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"withdrawn.v source *.n" )==1)) addnote2(sock,"departing","source",rnode);
 if((len==1)&& (matchlike(vec,"withdrawn.v source *.n" )==1)) addnote2(sock,"perception_noise","source",rnode);
 if((len==1)&& (matchlike(vec,"withdrawn.v source *.n" )==1)) addnote2(sock,"removing","source",rnode);
 if((len==1)&& (matchlike(vec,"withdrawn.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"withdrawn.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"wore.v object *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"works.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"worrying.v object *.n" )==1)) addnote2(sock,"awareness","content",rnode);
 /*
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"adornment","location",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"appearance","phenomenon",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"arriving","theme",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"cause-to-move","agent",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"cure","treatment",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"departing","theme",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"duplication","original",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"expectation","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"experiencer-obj","cause",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","content",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"motion_noise","theme",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"removing","theme",rnode);
 if((len==1)&& (matchlike(vec,"would.v theme *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 */
 if((len==1)&& (matchlike(vec,"wrap.v object *.n" )==1)) addnote2(sock,"filling","theme",rnode);
 if((len==1)&& (matchlike(vec,"wrap.v object *.n" )==1)) addnote2(sock,"placing","theme",rnode);
 if((len==1)&& (matchlike(vec,"wrinkled.v coagent-insturment *.n" )==1)) addnote2(sock,"body-movement","internal cause",rnode);
 if((len==1)&& (matchlike(vec,"yanked.v destination *.n" )==1)) addnote2(sock,"cause-to-move","goal",rnode);
 if((len==1)&& (matchlike(vec,"yanked.v object *.n" )==1)) addnote2(sock,"cause-to-move","theme",rnode);
 if((len==1)&& (matchlike(vec,"yawned.v contact *.n" )==1)) addnote2(sock,"posture","location",rnode);
 if((len==1)&& (matchlike(vec,"yearn.v theme *.n" )==1)) addnote2(sock,"experiencer-subj","experiencer",rnode);
 if((len==1)&& (matchlike(vec,"zigzagged.v theme *.n" )==1)) addnote2(sock,"path-shape","theme",rnode);
 /*
 if((len==1)&& (matchlike(vec,"`[?].v location *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"`[?].v location-time *.n" )==1)) addnote2(sock,"communication_noise","message",rnode);
 if((len==1)&& (matchlike(vec,"`[?].v location-time *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"`[?].v object *.n" )==1)) addnote2(sock,"awareness","cognizer",rnode);
 if((len==1)&& (matchlike(vec,"`[?].v object *.n" )==1)) addnote2(sock,"communication_noise","speaker",rnode);
 if((len==1)&& (matchlike(vec,"`[?].v object *.n" )==1)) addnote2(sock,"conversation","topic",rnode);
 if((len==1)&& (matchlike(vec,"`[?].v object *.n" )==1)) addnote2(sock,"questioning","message",rnode);
 if((len==1)&& (matchlike(vec,"`[?].v object *.n" )==1)) addnote2(sock,"statement","topic",rnode);
 if((len==1)&& (matchlike(vec,"`[?].v recipient *.n" )==1)) addnote2(sock,"manner","message",rnode);
 if((len==1)&& (matchlike(vec,"`[?].v theme *.n" )==1)) addnote2(sock,"separation","whole",rnode);
 */
}

 if (len==2)
 {
 if((len==2)&& (matchlike(vec,"arched.v theme one spec *.n" )==1)) addnote2(sock,"body-movement","body part",rnode);
 if((len==2)&& (matchlike(vec,"beckoned.v manner out recipient *.n" )==1)) addnote2(sock,"gesture","message",rnode);
 if((len==2)&& (matchlike(vec,"came.v destination one spec *.n" )==1)) addnote2(sock,"arriving","goal",rnode);
 if((len==2)&& (matchlike(vec,"get.v theme any spec *.n" )==1)) addnote2(sock,"body","experiencer",rnode);
 if((len==2)&& (matchlike(vec,"grinned.v manner up location *.n" )==1)) addnote2(sock,"body-movement","addressee",rnode);
 if((len==2)&& (matchlike(vec,"leapt.v theme one spec *.n" )==1)) addnote2(sock,"self-motion","self-mover",rnode);
 if((len==2)&& (matchlike(vec,"looked.v manner out location *.n" )==1)) addnote2(sock,"appearance","direction",rnode);
 if((len==2)&& (matchlike(vec,"looked.v manner out location *.n" )==1)) addnote2(sock,"perception_active","direction",rnode);
 if((len==2)&& (matchlike(vec,"overheard.v theme one spec *.n" )==1)) addnote2(sock,"hear","addressee",rnode);
 if((len==2)&& (matchlike(vec,"reproduced.v location-time 3 spec *.n" )==1)) addnote2(sock,"duplication","goal",rnode);
 if((len==2)&& (matchlike(vec,"rowed.v manner out destination *.n" )==1)) addnote2(sock,"transportation","goal",rnode);
 if((len==2)&& (matchlike(vec,"stared.v manner up location *.n" )==1)) addnote2(sock,"perception_active","phenomenon",rnode);
 if((len==2)&& (matchlike(vec,"swapped.v theme one spec *.n" )==1)) addnote2(sock,"replacement","themes",rnode);
 if((len==2)&& (matchlike(vec,"take.v object any spec *.n" )==1)) addnote2(sock,"becoming-aware","phenomenon",rnode);
 if((len==2)&& (matchlike(vec,"was.v object one spec *.n" )==1)) addnote2(sock,"departing","source",rnode);
 }

}




void cyc_inferlinks(SOCKET sock, Linkage linkage)
{
	cyclishQ=1;
    kb_buf[0]=0;
	kb_hbuf[0]=0;
	kb_tbuf[0]=0;
	kb_tempb[0]=0;
	kb_trigrams[0]=0;

	inferlinks_common(sock,linkage);
	kb_tempb[0]=0;
	sprintf(kb_tempb,"<cyc-hypolinks>\r\n%s (#$and %s ) %s \r\n</cyc-hypolinks>\r\n<trigrams>\r\n%s\r\n</trigrams>\r\n",kb_hbuf,kb_buf,kb_tbuf,kb_trigrams);
	OutputDebugString(kb_tempb);
	toCyc(kb_tempb);
	if (cycsafe(kb_tempb)==1) sock_print(sock,kb_tempb);

	cyclishQ=0;
}

void inferlinks(SOCKET sock, Linkage linkage)
{
	cyclishQ=0;
    kb_buf[0]=0;
	kb_hbuf[0]=0;
	kb_tbuf[0]=0;
	kb_tempb[0]=0;
	kb_trigrams[0]=0;
	inferlinks_common(sock,linkage);
	cyclishQ=0;
}
